//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_SDL_GUI.cpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#include "LaserBoy_SDL_GUI.hpp"

//############################################################################
static void SDL_bailout(int rc)
{
    SDL_Quit();
    exit(rc);
}

//############################################################################
LaserBoy_SDL_GUI::LaserBoy_SDL_GUI(int argc, char *argv[])
                : screen(NULL)
{
    if(argc >= 2)
    {
        xres = atoi(argv[1]);
        yres = atoi(argv[2]);
    }

    if(SDL_Init(SDL_INIT_VIDEO) < 0)
        SDL_bailout(2);
    if((screen = SDL_SetVideoMode(xres,
                                  yres,
                                  0,
                                    SDL_HWSURFACE
                                  | SDL_ANYFORMAT
                                 )
       ) == NULL
      )
        SDL_bailout(2);

    SDL_EnableUNICODE(1);
    SDL_EnableKeyRepeat(250, 1);

    space.font_size_factor = screen->h / 480;

    if(space.font_size_factor == 0)
        space.font_size_factor = 1;

    SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
    SDL_EventState(SDL_PRESSED    , SDL_IGNORE);
    SDL_EventState(SDL_RELEASED   , SDL_IGNORE);
    SDL_EventState(SDL_SYSWMEVENT , SDL_IGNORE);

    SDL_WM_SetCaption("LaserBoy Multiplatform by James Lehman : Extra Stimulus Inc.", NULL);

    if(SDL_LockSurface(screen) < 0)
        SDL_bailout(2);

    bmp_clone_SDL_Surface(space.p_bmp(), screen, 0);
    bmp_copy(space.p_background()    , space.p_bmp());
    bmp_copy(space.p_background_bmp(), space.p_bmp());

    space.recolor_background();
}

//############################################################################
LaserBoy_SDL_GUI::~LaserBoy_SDL_GUI()
{
    SDL_FreeSurface(screen);
    SDL_Quit       ();
}

//############################################################################
string LaserBoy_SDL_GUI::display_prompt_and_echo_name(string prompt)
{
    bool   done         = false;
    int    x_off        = LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor,
           y_off        = space.bmp.yres - 24 * space.font_size_factor,
           x_gap        = 8 * space.font_size_factor,
           y_line       = x_gap;
    u_int  bmp_bg_color = space.bmp.make_rgb(space.bg_color.r, space.bg_color.g, space.bg_color.b);
    string input;

    prompt_escaped = false;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.white,
               bmp_bg_color,
               0,
               space.font_size_factor,
               prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.make_rgb(0,255,255),
               bmp_bg_color,
               0,
               space.font_size_factor,
               "_"
              );

    display_space();

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            prompt_escaped = true;
            return "";
        }
        if(event.key.keysym.unicode == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(&(space.bmp),
                                   x_off,
                                   y_off,
                                   space.bmp.make_rgb(0,255,255),
                                   bmp_bg_color,
                                   0,
                                   space.font_size_factor,
                                   " "
                                  );
                        display_space();
                    }
                    break;
                //------------------------------------------------------------
                default:
                    if(    isalnum(event.key.keysym.unicode)
                        || event.key.keysym.unicode == '_'
                        || event.key.keysym.unicode == '.'
                      )
                    {
                        bmp_printf(&(space.bmp),
                                   x_off,
                                   y_off,
                                   space.bmp.make_rgb(0,255,255),
                                   bmp_bg_color,
                                   0,
                                   space.font_size_factor,
                                   "%c",
                                   event.key.keysym.unicode
                                  );
                        display_space();
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }
                    break;
            } // end switch(event.key.keysym.unicode)
        } // end else of if(event.key.keysym.unicode == SDLK_RETURN)
    } // end while(!done)
    bmp_printf(&(space.bmp),
               x_off - x_gap * input.size(),
               y_off,
               space.bmp.make_rgb(255,0,0),
               bmp_bg_color,
               0,
               space.font_size_factor,
               input.c_str()
              );
    display_space();
    return input;
}

//############################################################################
string LaserBoy_SDL_GUI::display_prompt_and_echo_string(string prompt)
{
    bool   done         = false;
    int    x_off        = LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor,
           y_off        = space.bmp.yres - 24 * space.font_size_factor,
           x_gap        = 8 * space.font_size_factor,
           y_line       = 8 * space.font_size_factor;
    u_int  bmp_bg_color = space.bmp.make_rgb(space.bg_color.r, space.bg_color.g, space.bg_color.b);
    string input;

    prompt_escaped = false;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.white,
               bmp_bg_color,
               0,
               space.font_size_factor,
               prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.make_rgb(0,255,255),bmp_bg_color,
               0,
               space.font_size_factor,
               "_"
              );
    display_space();

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            prompt_escaped = true;
            return "";
        }
        if(event.key.keysym.unicode == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(&(space.bmp),
                                   x_off,
                                   y_off,
                                   space.bmp.make_rgb(0,255,255),
                                   bmp_bg_color,
                                   0,
                                   space.font_size_factor,
                                   " "
                                  );
                        display_space();
                    }
                    break;
                //------------------------------------------------------------
                default:
                    if(    isgraph(event.key.keysym.unicode)
                        || event.key.keysym.unicode == ' '
                      )
                    {
                        bmp_printf(&(space.bmp),
                                   x_off,
                                   y_off,
                                   space.bmp.make_rgb(0,255,255),
                                   bmp_bg_color,
                                   0,
                                   space.font_size_factor,
                                   "%c",
                                   event.key.keysym.unicode
                                  );
                        display_space();
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }
                    break;
            } // end switch(event.key.keysym.unicode)
        } // end else of if(event.key.keysym.unicode == SDLK_RETURN)
    } // end while(!done)
    bmp_printf(&(space.bmp),
               x_off - x_gap * input.size(),
               y_off,
               space.bmp.make_rgb(255,0,0),
               bmp_bg_color,
               0,
               space.font_size_factor,
               input.c_str()
              );
    display_space();
    return input;
}

//############################################################################
double LaserBoy_SDL_GUI::display_prompt_and_echo_double(string prompt)
{
    bool   done         = false;
    int    x_off        = LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor,
           y_off        = space.bmp.yres - 24 * space.font_size_factor,
           x_gap        = 8 * space.font_size_factor,
           y_line       = 8 * space.font_size_factor;
    u_int  bmp_bg_color = space.bmp.make_rgb(space.bg_color.r, space.bg_color.g, space.bg_color.b);
    double result;
    string input;

    prompt_escaped = false;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.white,
               bmp_bg_color,
               0,
               space.font_size_factor,
               prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.make_rgb(0,255,255),
               bmp_bg_color,
               0,
               space.font_size_factor,
               "_"
              );
    display_space();

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            prompt_escaped = true;
            return 0.0;
        }
        if(event.key.keysym.unicode == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(&(space.bmp),
                                   x_off,
                                   y_off,
                                   space.bmp.make_rgb(0,255,255),
                                   bmp_bg_color,
                                   0,
                                   space.font_size_factor,
                                   " "
                                  );
                        display_space();
                    }
                    break;
                //------------------------------------------------------------
                default:
                    if(    isdigit(event.key.keysym.unicode)
                        || (    event.key.keysym.unicode == '.'
                             && input.find('.') == string::npos
                           )
                        || (    event.key.keysym.unicode == '+'
                             && !input.size()
                           )
                        || (    event.key.keysym.unicode == '-'
                             && !input.size()
                           )
                      )
                    {
                        bmp_printf(&(space.bmp),
                                   x_off,
                                   y_off,
                                   space.bmp.make_rgb(0,255,255),
                                   bmp_bg_color,
                                   0,
                                   space.font_size_factor,
                                   "%c",
                                   event.key.keysym.unicode
                                  );
                        display_space();
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }
                    break;
            } // end switch(event.key.keysym.unicode)
        } // end else of if(event.key.keysym.unicode == SDLK_RETURN)
    } // end while(!done)
    bmp_printf(&(space.bmp),
               x_off - x_gap * input.size(),
               y_off,
               space.bmp.make_rgb(255,0,0),
               bmp_bg_color,
               0,
               space.font_size_factor,
               input.c_str()
              );
    display_space();
    if(input.size())
        sscanf(input.c_str(), "%lf", &result);
    else
        result = 0.0;
    return result;
}

//############################################################################
int LaserBoy_SDL_GUI::display_prompt_and_echo_int(string prompt)
{
    bool   done         = false;
    int    x_off        = LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor,
           y_off        = space.bmp.yres - 24 * space.font_size_factor,
           x_gap        = 8 * space.font_size_factor,
           y_line       = 8 * space.font_size_factor,
           result;
    u_int  bmp_bg_color = space.bmp.make_rgb(space.bg_color.r, space.bg_color.g, space.bg_color.b);
    string input;

    prompt_escaped = false;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.white,
               bmp_bg_color,
               0,
               space.font_size_factor,
               prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.make_rgb(0,255,255),
               bmp_bg_color,
               0,
               space.font_size_factor,
               "_"
              );
    display_space();

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            prompt_escaped = true;
            return 0;
        }
        if(event.key.keysym.unicode == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(&(space.bmp),
                                   x_off,
                                   y_off,
                                   space.bmp.make_rgb(0,255,255),
                                   bmp_bg_color,
                                   0,
                                   space.font_size_factor,
                                   " "
                                  );
                        display_space();
                    }
                    break;
                //------------------------------------------------------------
                default:
                    if(    isdigit(event.key.keysym.unicode)
                        || (    event.key.keysym.unicode == '+'
                             && !input.size()
                           )
                        || (    event.key.keysym.unicode == '-'
                             && !input.size()
                           )
                      )
                    {
                        bmp_printf(&(space.bmp),
                                   x_off,
                                   y_off,
                                   space.bmp.make_rgb(0,255,255),
                                   bmp_bg_color,
                                   0,
                                   space.font_size_factor,
                                   "%c",
                                   event.key.keysym.unicode
                                  );
                        display_space();
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }
                    break;
            } // end switch(event.key.keysym.unicode)
        } // end else of if(event.key.keysym.unicode == SDLK_RETURN)
    } // end while(!done)
    bmp_printf(&(space.bmp),
               x_off - x_gap * input.size(),
               y_off,
               space.bmp.make_rgb(255,0,0),
               bmp_bg_color,
               0,
               space.font_size_factor,
               input.c_str()
              );
    display_space();
    if(input.size())
        sscanf(input.c_str(), "%d", &result);
    else
        result = 0;
    return result;
}

//############################################################################
int LaserBoy_SDL_GUI::display_prompt_and_echo_nibble(string prompt)
{
    bool   done         = false;
    int    x_off        = LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor,
           y_off        = space.bmp.yres - 24 * space.font_size_factor,
           y_line       = 8 * space.font_size_factor,
           result;
    u_int  bmp_bg_color = space.bmp.make_rgb(space.bg_color.r, space.bg_color.g, space.bg_color.b);

    prompt_escaped = false;

    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.white,
               bmp_bg_color,
               0,
               space.font_size_factor,
               prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(&(space.bmp),
               x_off,
               y_off,
               space.bmp.make_rgb(0,255,255),
               bmp_bg_color,
               0,
               space.font_size_factor,
               "_"
              );
    display_space();

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);
        if(event.key.keysym.unicode == SDLK_ESCAPE)
        {
            prompt_escaped = true;
            return -1;
        }
        switch(event.key.keysym.unicode)
        {
            case '0':    return 0x00;
            case '1':    return 0x01;
            case '2':    return 0x02;
            case '3':    return 0x03;
            case '4':    return 0x04;
            case '5':    return 0x05;
            case '6':    return 0x06;
            case '7':    return 0x07;
            case '8':    return 0x08;
            case '9':    return 0x09;
            case 'a':    return 0x0a;
            case 'b':    return 0x0b;
            case 'c':    return 0x0c;
            case 'd':    return 0x0d;
            case 'e':    return 0x0e;
            case 'f':    return 0x0f;
            //----------------------------------------------------------------
        } // end switch(event.key.keysym.unicode)
    } // end while(!done)
    return result;
}

//############################################################################
bool LaserBoy_SDL_GUI::display_prompt_and_echo_bool(string prompt)
{
    int    x_off  = LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor,
           y_off  = space.bmp.yres - 24 * space.font_size_factor;

    prompt_escaped = false;
    bmp_printf(   &(space.bmp)
                , x_off
                , y_off
                , space.bmp.white
                , space.bmp.make_rgb(space.bg_color.r, space.bg_color.g, space.bg_color.b)
                , 0
                , space.font_size_factor
                , prompt.c_str()
              );
    display_space();
    do
    {
        SDL_WaitEvent(&event);
    } while(event.type != SDL_KEYDOWN);
    if(event.key.keysym.unicode == SDLK_ESCAPE)
    {
        prompt_escaped = true;
        return false;
    }
    if(    event.key.keysym.unicode == 'y'
        || event.key.keysym.unicode == '1'
      )
        return true;
    return false;
}

//############################################################################
void LaserBoy_SDL_GUI::display_error(string err)
{
    space.TUI_clue = "oops!";
    space.render_space();
    //------------------------------------------------------------------------
    bmp_printf(   &(space.bmp)
                , LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor
                , LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor
                , space.bmp.white
                , space.bmp.make_rgb(192, 0, 0)
                , 0
                , space.font_size_factor
                , "%s                                                                                                                         "
                , err.c_str()
              );
    //------------------------------------------------------------------------
    bmp_printf(   &(space.bmp)
                , LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor
                , space.bmp.yres - 24 * space.font_size_factor
                , space.bmp.white
                , space.bmp.make_rgb(192, 0, 0)
                , 0
                , space.font_size_factor
                , "[any] key to continue"
              );
    //------------------------------------------------------------------------
    display_space();
    return ;
}

//############################################################################
void LaserBoy_SDL_GUI::wait_4_key()
{
    while(true)
    {
        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
            break;
    }
    return;
}

//############################################################################
// wave functions that require direct access to space.bmp
//############################################################################
bool LaserBoy_SDL_GUI::display_wave(string file_name)
{
    ifstream  in;
    do // clear key strokes from event queue!
    {
        SDL_PollEvent(&event);
    } while(event.type == SDL_KEYDOWN);

    in.open(file_name.c_str(), ios::in | ios::binary);
    if(!in.is_open())
        return false;

    LaserBoy_wave_header header(in);
    if(    header.version == "!LaserBoy!"
        || !(header.LaserBoy_wave_mode & LASERBOY_WAVE_END_OF_FRAME)
      )
        display_unframed_wave(in, header);
    else
        display_LaserBoy_wave(in, header);
    in.close();
    return true;
}

//############################################################################
bool LaserBoy_SDL_GUI::display_unframed_wave(string file_name)
{
    ifstream in;
    do // clear key strokes from event queue!
    {
        SDL_PollEvent(&event);
    } while(event.type == SDL_KEYDOWN);

    in.open(file_name.c_str(), ios::in | ios::binary);
    if(!in.is_open())
        return false;

    LaserBoy_wave_header header(in);
    display_unframed_wave(in, header);
    in.close();
    return true;
}

//############################################################################
void LaserBoy_SDL_GUI::display_LaserBoy_wave(ifstream& in, LaserBoy_wave_header header)
{
    bool                     playing      = true         ;

    int                      i                           ,
                             error                       ,
                             span                        ,
                             last                        ,
                             red                         ,
                             green                       ,
                             blue                        ,
                             sample_count = 0            ,
                             offset_max                  ,
                             offset_min                  ,
                             center       = screen->h / 2;

    u_int                    time_zero                   ,
                             color        = space.bmp.white,
                             bmp_bg_color = space.bmp.make_rgb(space.bg_color.r,
                                                               space.bg_color.g,
                                                               space.bg_color.b
                                                              );
    double                  scale  = screen->h / 65536.0;
    struct  LaserBoy_bmp    frame_buffer = {0}          ;

    LaserBoy_wave_sample_pointer   roll_over    ;
    LaserBoy_wave_sample_pointer*  sample_window;
    //------------------------------------------------------------------------
    if(header.num_channels < 6)
    {
        bmp_printf(&(space.bmp),
                   LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor,
                   LASERBOY_STATS_DISPLAY_OFFSET * space.font_size_factor,
                   space.bmp.white,
                   bmp_bg_color,
                   0,
                   space.font_size_factor,
                   "[any key] this is not a LaserBoy wave        "
                  );
        error = LASERBOY_INCOMPATIBLE_FORMATS;
        display_space();
        prompt_escaped = true;
        return;
    }
    //------------------------------------------------------------------------
    if(!header.LaserBoy_wave_mode & LASERBOY_WAVE_OFFSETS) // if no offsets in header
        for(i = 0; i < 6; i++)
            header.offset[i] = 0; // give me some!
    //------------------------------------------------------------------------
    offset_min = header.offset[0];
    offset_max = header.offset[0];
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    //------------------------------------------------------------------------
    span  = (offset_max - offset_min) + 2; // at least 2 samples!
    last  = span - 1;
    red   = -header.offset[2] + 1;
    green = -header.offset[3] + 1;
    blue  = -header.offset[4] + 1;
    //------------------------------------------------------------------------
    sample_window = new LaserBoy_wave_sample_pointer[span]; // at least 2
    for(i = 0; i < span; i++)
        sample_window[i] = new LaserBoy_wave_sample(header.num_channels);
    //------------------------------------------------------------------------
    bmp_clone_SDL_Surface(&frame_buffer, screen, 0);
    bmp_uint(&frame_buffer, space.bg_color.as_uint());
    //------------------------------------------------------------------------
    while(playing)
    {
        time_zero = SDL_GetTicks();
        //--------------------------------------------------------------------
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;
        //--------------------------------------------------------------------
        if(space.wave_show_inverted)
            playing = sample_window[last]->from_ifstream_wave_inverted(in);
        else
            playing = sample_window[last]->from_ifstream_wave(in);
        //--------------------------------------------------------------------
        sample_count++;
        //--------------------------------------------------------------------
        color = frame_buffer.make_rgb(   sample_window[red  ]->channel[2] >> 7
                                       , sample_window[green]->channel[3] >> 7
                                       , sample_window[blue ]->channel[4] >> 7
                                     );
        //--------------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
            color = 0;
        //--------------------------------------------------------------------
        if(color)
        {
            bmp_put_line_fast(   &frame_buffer
                               , (u_int)((sample_window[0]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[0]->channel[1] * -scale) + center)
                               , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                               , color
                             );

            if(space.wave_show_vertices)
                bmp_put_vertex(   &frame_buffer
                                , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                                , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                                , color
                              );
        }
        //--------------------------------------------------------------------
        else if(space.wave_show_blanks)
        {
            bmp_put_line_fast(   &frame_buffer
                               , (u_int)((sample_window[0]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[0]->channel[1] * -scale) + center)
                               , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                               , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                               , frame_buffer.gray
                             );

            if(space.wave_show_vertices)
                bmp_put_vertex(   &frame_buffer
                                , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                                , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                                , frame_buffer.gray
                              );
        }
        //----------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
        {
            if(space.wave_real_time)
            {
                do
                {
                     SDL_Delay(1);
                }
                while(   SDL_GetTicks() - time_zero
                       < (u_int)(sample_count / (header.sample_rate / 1000.0))
                     );
            }

            bmp_to_SDL_Surface(&frame_buffer, screen);

            if(space.wave_clean_screen)
                bmp_uint(&frame_buffer, space.bg_color.as_uint());

            SDL_PollEvent(&event);

            if(event.type == SDL_KEYDOWN)
                playing = false;

            sample_count = 0;
        }
        //--------------------------------------------------------------------
    } // end while(playing)
    bmp_free(&frame_buffer);
    delete sample_window;
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::display_unframed_wave(ifstream& in, LaserBoy_wave_header header)
{
    bool                     playing      = true         ;

    int                      i                           ,
                             error                       ,
                             span                        ,
                             last                        ,
                             red                         ,
                             green                       ,
                             blue                        ,
                             sample_count = 0            ,
                             offset_max                  ,
                             offset_min                  ,
                             center       = screen->h / 2;

    u_int                    time_zero                     ,
                             color        = space.bmp.white,
                             bmp_bg_color = space.bmp.make_rgb(space.bg_color.r,
                                                               space.bg_color.g,
                                                               space.bg_color.b
                                                              );

    double                   scale  = screen->h / 65536.0;
    struct  LaserBoy_bmp     frame_buffer = {0}          ;

    LaserBoy_wave_sample_pointer   roll_over    ;
    LaserBoy_wave_sample_pointer*  sample_window;
    //------------------------------------------------------------------------
    offset_max = space.wave_offsets[0];
    for(i = 1; i < 6; i++)
        if(space.wave_offsets[i] > offset_max)
            offset_max = space.wave_offsets[i];
    //------------------------------------------------------------------------
    for(i = 0; i < 6; i++)
        space.wave_offsets[i] -= offset_max;
    //------------------------------------------------------------------------
    for(i = 0; i < 6; i++)
        header.offset[i] = space.wave_offsets[i]; // use current system offsets
    //------------------------------------------------------------------------
    offset_min = header.offset[0];
    offset_max = header.offset[0];
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    //------------------------------------------------------------------------
    span  = (offset_max - offset_min) + 2; // at least 2 samples!
    last  = span - 1;
    red   = -header.offset[2] + 1;
    green = -header.offset[3] + 1;
    blue  = -header.offset[4] + 1;
    //------------------------------------------------------------------------
    sample_window = new LaserBoy_wave_sample_pointer[span]; // at least 2
    for(i = 0; i < span; i++)
        sample_window[i] = new LaserBoy_wave_sample(header.num_channels);
    //------------------------------------------------------------------------
    bmp_clone_SDL_Surface(&frame_buffer, screen, 0);
    bmp_uint(&frame_buffer, space.bg_color.as_uint());
    //------------------------------------------------------------------------
    while(playing)
    {
        time_zero = SDL_GetTicks();
        //--------------------------------------------------------------------
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;
        //--------------------------------------------------------------------
        if(space.wave_show_inverted)
            playing = sample_window[last]->from_ifstream_wave_inverted(in);
        else
            playing = sample_window[last]->from_ifstream_wave(in);
        //--------------------------------------------------------------------
        sample_count++;
        //--------------------------------------------------------------------
        if(header.num_channels >= 6)
            color = frame_buffer.make_rgb(   (   (sample_window[red  ]->channel[2] > 0)
                                               ? (sample_window[red  ]->channel[2] / 128)
                                               : (0)
                                             )
                                           , (   (sample_window[green]->channel[3] > 0)
                                               ? (sample_window[green]->channel[3] / 128)
                                               : (0)
                                             )
                                           , (   (sample_window[blue ]->channel[4] > 0)
                                               ? (sample_window[blue ]->channel[4] / 128)
                                               : (0)
                                             )
                                         );
        //--------------------------------------------------------------------
        else if(header.num_channels >= 4)
            color = frame_buffer.make_rgb(   (   (sample_window[red  ]->channel[2] > 0)
                                               ? (sample_window[red  ]->channel[2] / 128)
                                               : (0)
                                             )
                                           , (   (sample_window[red  ]->channel[2] > 0)
                                               ? (sample_window[red  ]->channel[2] / 128)
                                               : (0)
                                             )
                                           , (   (sample_window[red  ]->channel[2] > 0)
                                               ? (sample_window[red  ]->channel[2] / 128)
                                               : (0)
                                             )
                                         );
        //--------------------------------------------------------------------
        else
            color = frame_buffer.make_rgb(255, 255, 255);
        //--------------------------------------------------------------------
        bmp_put_line_fast(   &frame_buffer
                           , (u_int)((sample_window[0]->channel[0] *  scale) + center)
                           , (u_int)((sample_window[0]->channel[1] * -scale) + center)
                           , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                           , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                           , color
                         );

        if(space.wave_show_vertices)
            bmp_put_vertex(   &frame_buffer
                            , (u_int)((sample_window[1]->channel[0] *  scale) + center)
                            , (u_int)((sample_window[1]->channel[1] * -scale) + center)
                            , color
                          );
        //--------------------------------------------------------------------
        if(!(sample_count % space.unframed_wave_stride))
        {
            if(space.wave_real_time)
            {
                do
                {
                     SDL_Delay(1);
                }
                while(   SDL_GetTicks() - time_zero
                       < (u_int)(sample_count / (header.sample_rate / 1000.0))
                     );
            }

            bmp_to_SDL_Surface(&frame_buffer, screen);

            if(space.wave_clean_screen)
                bmp_uint(&frame_buffer, space.bg_color.as_uint());
            SDL_PollEvent(&event);
            if(event.type == SDL_KEYDOWN)
                playing = false;
            sample_count = 0;
        }
        //--------------------------------------------------------------------
    } // end while(playing)
    bmp_free(&frame_buffer);
    delete sample_window;
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::check_for_SDL_arrow_keys(SDLKey key)
{
    switch(key)
    {
        default:
            space.TUI_clue = "bad key";
            break;
        //--------------------------------------------------------------------
        case SDLK_UP: // show next palette
            space.target_palette_index++;
            if(space.target_palette_index >= space.number_of_palettes())
                space.target_palette_index = 0;
            space.TUI_clue = "+ target palette";
            break;
        //--------------------------------------------------------------------
        case SDLK_DOWN: // show previous palette
            space.target_palette_index--;
            if(space.target_palette_index < 0)
                space.target_palette_index = space.number_of_palettes() - 1;
            space.TUI_clue = "- target palette";
            break;
        //--------------------------------------------------------------------
        case SDLK_RIGHT: // show previous frame
            if(    space.loop_animation
                || (space.frame_index < space.number_of_frames() - 1)
              )
                ++space;
            space.TUI_clue = "+ frame";
            break;
        //--------------------------------------------------------------------
        case SDLK_LEFT: // show next frame
            if(    space.loop_animation
                || (space.frame_index > 0)
              )
                --space;
            space.TUI_clue = "- frame";
            break;
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::animate_forward()
{
    u_int  time_zero;

    SDL_Delay(250);
    SDL_PollEvent(&event);

    space.TUI_clue = "+ show";
    while(    (event.type != SDL_KEYDOWN)
           && (    space.loop_animation
                || (space.frame_index < space.number_of_frames() - 1)
              )
         )
    {
        time_zero = SDL_GetTicks();
        ++space;
        space.render_space();
        display_space();
        if(space.approximate_frame_rate)
        {
            do{;}
            while(   SDL_GetTicks() - time_zero
                   < (u_int)((1.0 / space.frames_per_second) * 1000.0)
                 );
        }
        SDL_PollEvent(&event);
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::animate_reverse()
{
    u_int  time_zero;

    SDL_Delay(250);
    SDL_PollEvent(&event);

    space.TUI_clue = "- show";

    while(    (event.type != SDL_KEYDOWN)
           && (    space.loop_animation
                || (space.frame_index > 0)
              )
         )
    {
        time_zero = SDL_GetTicks();
        --space;
        space.render_space();
        display_space();
        if(space.approximate_frame_rate)
        {
            do{;}
            while(   SDL_GetTicks() - time_zero
                   < (u_int)((1.0 / space.frames_per_second) * 1000.0)
                 );
        }
        SDL_PollEvent(&event);
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::rotate_forward()
{
    u_int  time_zero;

    SDL_Delay(250);
    SDL_PollEvent(&event);

    space.TUI_clue = "+ orbit";

    while(    (event.type != SDL_KEYDOWN)
           && (    space.loop_animation
                || (space.frame_index < space.number_of_frames() - 1)
              )
         )
    {
        time_zero = SDL_GetTicks();
        space.view_angle += one_degree;
        space.render_space();
        display_space();
        if(space.approximate_frame_rate)
        {
            do{;}
            while(   SDL_GetTicks() - time_zero
                   < (u_int)((1.0 / space.frames_per_second) * 1000.0)
                 );
        }
        SDL_PollEvent(&event);
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::rotate_reverse()
{
    u_int  time_zero;

    SDL_Delay(250);
    SDL_PollEvent(&event);

    space.TUI_clue = "- orbit";

    while(    (event.type != SDL_KEYDOWN)
           && (    space.loop_animation
                || (space.frame_index > 0)
              )
         )
    {
        time_zero = SDL_GetTicks();
        space.view_angle -= one_degree;
        space.render_space();
        display_space();
        if(space.approximate_frame_rate)
        {
            do{;}
            while(   SDL_GetTicks() - time_zero
                   < (u_int)((1.0 / space.frames_per_second) * 1000.0)
                 );
        }
        SDL_PollEvent(&event);
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::animate_rotate_forward()
{
    u_int  time_zero;

    SDL_Delay(250);
    SDL_PollEvent(&event);

    space.TUI_clue = "+ show & orbit";

    while(    (event.type != SDL_KEYDOWN)
           && (    space.loop_animation
                || (space.frame_index < space.number_of_frames() - 1)
              )
         )
    {
        time_zero = SDL_GetTicks();
        ++space;
        space.view_angle += one_degree;
        space.render_space();
        display_space();
        if(space.approximate_frame_rate)
        {
            do{;}
            while(   SDL_GetTicks() - time_zero
                   < (u_int)((1.0 / space.frames_per_second) * 1000.0)
                 );
        }
        SDL_PollEvent(&event);
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::animate_rotate_reverse()
{
    u_int  time_zero;

    SDL_Delay(250);
    SDL_PollEvent(&event);

    space.TUI_clue = "- show & orbit";

    while(    (event.type != SDL_KEYDOWN)
           && (    space.loop_animation
                || (space.frame_index > 0)
              )
         )
    {
        time_zero = SDL_GetTicks();
        --space;
        space.view_angle -= one_degree;
        space.render_space();
        display_space();
        if(space.approximate_frame_rate)
        {
            do{;}
            while(   SDL_GetTicks() - time_zero
                   < (u_int)((1.0 / space.frames_per_second) * 1000.0)
                 );
        }
        SDL_PollEvent(&event);
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::system_settings_menu()
{
    bool   in_menu = true;
    int    temp_int;
    double temp_double;

    space.TUI_clue = "program settings";

    while(in_menu)
    {
        space.render_space();
        space.overlay_settings_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    in_menu = false;
                    space.TUI_clue = "exit settings";
                    break;
                //------------------------------------------------------------
                case '1':
                    temp_double = (display_prompt_and_echo_double("enter new insignificant angle in degrees") * one_degree);
                    if(prompt_escaped)
                        break;
                    space.insignificant_angle = temp_double;
                    break;
                //------------------------------------------------------------
                case '2':
                    temp_double = display_prompt_and_echo_double("enter insignificant distance in ordinates");
                    if(prompt_escaped)
                        break;
                    space.insignificant_distance = temp_double;
                    break;
                //------------------------------------------------------------
                case '3':
                    temp_double = display_prompt_and_echo_double("enter new maximum lit vector length in ordinates");
                    if(prompt_escaped)
                        break;
                    space.lit_delta_max = temp_double;
                    break;
                //------------------------------------------------------------
                case '4':
                    temp_double = display_prompt_and_echo_double("enter new maximum blank vector length in ordinates");
                    if(prompt_escaped)
                        break;
                    space.blank_delta_max = temp_double;
                    break;
                //------------------------------------------------------------
                case '5':
                    temp_double = display_prompt_and_echo_double("enter dwell in microseconds for 180 degree angle");
                    if(prompt_escaped)
                        break;
                    space.max_dwell_microsec = temp_double;
                    break;
                //------------------------------------------------------------
                case '6':
                    temp_int = display_prompt_and_echo_int("enter new lit dwell overhang");
                    if(prompt_escaped)
                        break;
                    space.lit_dwell_overhang = temp_int;
                    break;
                //------------------------------------------------------------
                case '7':
                    dots_setting_menu();
                    break;
                //------------------------------------------------------------
                case '8':
                    temp_int = display_prompt_and_echo_int("enter new dwell on dot");
                    if(prompt_escaped)
                        break;
                    space.dwell_on_dot = temp_int;
                    break;
                //------------------------------------------------------------
                case '9':
                    temp_int = display_prompt_and_echo_int("enter new move ordinates per key");
                    if(prompt_escaped)
                        break;
                    space.displacement_step = temp_int;
                    break;
                //------------------------------------------------------------
                case '0':
                    temp_double = display_prompt_and_echo_double("rotate degrees per key hit");
                    if(prompt_escaped)
                        break;
                    space.rotation_step = temp_double * one_degree;
                    break;
                //------------------------------------------------------------
                case 'a':
                    temp_double = display_prompt_and_echo_double("scale percent per key hit");
                    if(prompt_escaped)
                        break;
                    space.magnitude_step = temp_double / 100.00;
                    space.magnitude_step_up = 1.00 + space.magnitude_step;
                    space.magnitude_step_dn = 1.00 - space.magnitude_step;
                    break;
                //------------------------------------------------------------
                case 'b':
                    temp_double = display_prompt_and_echo_double("rendered arc segment angle in degrees");
                    if(prompt_escaped)
                        break;
                    space.rendered_arc_angle = temp_double;
                    break;
                //------------------------------------------------------------
                case 'c':
                    temp_double = display_prompt_and_echo_double("enter new frames per second");
                    if(prompt_escaped)
                        break;
                    space.frames_per_second = temp_double;
                    break;
                //------------------------------------------------------------
                case 'd':
                    temp_double = display_prompt_and_echo_double("enter new wave frame still output durration in seconds");
                    if(prompt_escaped)
                        break;
                    space.still_frame_duration = temp_double;
                    break;
                //------------------------------------------------------------
                case 'e':
                    temp_int = display_prompt_and_echo_int("enter new output wave sample rate");
                    if(prompt_escaped)
                        break;
                    space.sample_rate = temp_int;
                    break;
                //------------------------------------------------------------
                case 'f':
                    temp_int = display_prompt_and_echo_int("enter new unframed wave samples per frame");
                    if(prompt_escaped)
                        break;
                    space.unframed_wave_stride = temp_int;
                    break;
                //------------------------------------------------------------
                case 'g':
                    signal_if_not_z_menu();
                    break;
                //------------------------------------------------------------
                case 'h':
                    temp_int = display_prompt_and_echo_int("output bitmap size in pixels square");
                    if(prompt_escaped)
                        break;
                    space.output_bmp_size = temp_int;
                    break;
                //------------------------------------------------------------
                case 'i':
                    temp_int = display_prompt_and_echo_int("screen fonst size factor");
                    if(prompt_escaped)
                        break;
                    space.font_size_factor = temp_int;
                    break;
                //------------------------------------------------------------
                //------------------------------------------------------------
                case 'j':
                    temp_int = display_prompt_and_echo_int("screen background color red");
                    if(prompt_escaped)
                        break;
                    space.bg_color.r = temp_int;
                    space.recolor_background();
                    break;
                //------------------------------------------------------------
                case 'k':
                    temp_int = display_prompt_and_echo_int("screen background color green");
                    if(prompt_escaped)
                        break;
                    space.bg_color.g = temp_int;
                    space.recolor_background();
                    break;
                //------------------------------------------------------------
                case 'l':
                    temp_int = display_prompt_and_echo_int("screen background color blue");
                    if(prompt_escaped)
                        break;
                    space.bg_color.b = temp_int;
                    space.recolor_background();
                    break;
                //------------------------------------------------------------
                case 'm':
                    temp_int = display_prompt_and_echo_int("screen background gray scale");
                    if(prompt_escaped)
                        break;
                    space.bg_color = (u_char)temp_int;
                    space.recolor_background();
                    break;
                //------------------------------------------------------------
                //------------------------------------------------------------
                case 'n':
                    temp_double = display_prompt_and_echo_double("enter new recolor palette span factor");
                    if(prompt_escaped)
                        break;
                    space.recolor_span_factor = temp_double;
                    break;
                //------------------------------------------------------------
                case 'o':
                    temp_int = display_prompt_and_echo_int("enter new black level");
                    if(prompt_escaped)
                        break;
                    space.black_level = (u_char)temp_int;
                    break;
            }
        }
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::signal_if_not_z_menu()
{
    space.TUI_clue = "signal if not Z";

    bool bad_key = true;
    while(bad_key)
    {
        space.render_space();
        space.overlay_if_not_z_options();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    space.TUI_clue = "action canceled";
                    bad_key = false;
                    break;
                //------------------------------------------------------------
                case '1':
                    space.signal_if_not_z = LASERBOY_SIGNAL_NO_SIGNAL;
                    bad_key = false;
                    break;
                //------------------------------------------------------------
                case '2':
                    space.signal_if_not_z = LASERBOY_SIGNAL_MONO_TTL;
                    bad_key = false;
                    break;
                //------------------------------------------------------------
                case '3':
                    space.signal_if_not_z = LASERBOY_SIGNAL_MONO_AVG_ANALOG;
                    bad_key = false;
                    break;
                //------------------------------------------------------------
                case '4':
                    space.signal_if_not_z = LASERBOY_SIGNAL_MONO_OR_ANALOG;
                    bad_key = false;
                    break;
                //------------------------------------------------------------
                case '5':
                    space.signal_if_not_z = LASERBOY_SIGNAL_MONO_WEIGHTED_ANALOG;
                    bad_key = false;
                    break;
            }
        }
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::dots_setting_menu()
{
    space.TUI_clue = "dots setting";

    bool bad_key = true;
    while(bad_key)
    {
        space.render_space();
        space.overlay_dots_options();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    space.TUI_clue = "action canceled";
                    bad_key = false;
                    break;
                //------------------------------------------------------------
                case '1':
                    space.dots_setting = LASERBOY_DOTS_IGNORE;
                    bad_key = false;
                    break;
                //------------------------------------------------------------
                case '2':
                    space.dots_setting = LASERBOY_DOTS_REMOVE;
                    bad_key = false;
                    break;
                //------------------------------------------------------------
                case '3':
                    space.dots_setting = LASERBOY_DOTS_ENHANCE;
                    bad_key = false;
                    break;
            }
        }
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::signal_bit_resolution()
{
    int temp_int;
    space.TUI_clue = "bit resolution";
    do
    {
        space.render_space();
        space.overlay_bit_resolution_settings();
        display_space();

        temp_int = display_prompt_and_echo_nibble("enter bits to mask (0 ~ f) for X");
        if(prompt_escaped)
            break;
        space.signal_bit_mask[0] = temp_int;

        space.render_space();
        space.overlay_bit_resolution_settings();
        display_space();
        temp_int = display_prompt_and_echo_nibble("enter bits to mask (0 ~ f) for Y");
        if(prompt_escaped)
            break;
        space.signal_bit_mask[1] = temp_int;

        space.render_space();
        space.overlay_bit_resolution_settings();
        display_space();
        temp_int = display_prompt_and_echo_nibble("enter bits to mask (0 ~ f) for Z or other");
        if(prompt_escaped)
            break;
        space.signal_bit_mask[5] = temp_int;

        space.render_space();
        space.overlay_bit_resolution_settings();
        display_space();
        temp_int = display_prompt_and_echo_nibble("enter bits to mask (0 ~ f) for red");
        if(prompt_escaped)
            break;
        space.signal_bit_mask[2] = temp_int;

        space.render_space();
        space.overlay_bit_resolution_settings();
        display_space();
        temp_int = display_prompt_and_echo_nibble("enter bits to mask (0 ~ f) for green");
        if(prompt_escaped)
            break;
        space.signal_bit_mask[3] = temp_int;

        space.render_space();
        space.overlay_bit_resolution_settings();
        display_space();
        temp_int = display_prompt_and_echo_nibble("enter bits to mask (0 ~ f) for blue");
        if(prompt_escaped)
            break;
        space.signal_bit_mask[4] = temp_int;

        space.render_space();
        space.overlay_bit_resolution_settings();
        display_space();
        temp_int = display_prompt_and_echo_nibble("enter bits to mask (0 ~ f) for L audio");
        if(prompt_escaped)
            break;
        space.signal_bit_mask[6] = temp_int;

        space.render_space();
        space.overlay_bit_resolution_settings();
        display_space();
        temp_int = display_prompt_and_echo_nibble("enter bits to mask (0 ~ f) for R audio");
        if(prompt_escaped)
            break;
        space.signal_bit_mask[7] = temp_int;
    } while(0);
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::visuals_settings_menu()
{
    bool in_menu = true;

    space.TUI_clue = "UI visuals";

    while(in_menu)
    {
        space.render_space();
        space.overlay_visuals_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                default:
                    space.TUI_clue = "bad key";
                    break;
                //------------------------------------------------------------
                case SDLK_ESCAPE:
                    in_menu = false;
                    space.TUI_clue = "exit UI visuals";
                    break;
                //------------------------------------------------------------
                case '1':
                    space.show_vertices = !space.show_vertices;
                    if(space.show_vertices)
                        space.TUI_clue = "show vertices";
                    else
                        space.TUI_clue = "hide vertices";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.show_blanking = !space.show_blanking;
                    if(space.show_blanking)
                        space.TUI_clue = "show blanking";
                    else
                        space.TUI_clue = "hide blanking";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.show_palettes = !space.show_palettes;
                    if(space.show_palettes)
                        space.TUI_clue = "show palettes";
                    else
                        space.TUI_clue = "hide palettes";
                    break;
                //------------------------------------------------------------
                case '4':
                    space.show_stats = !space.show_stats;
                    if(space.show_stats)
                        space.TUI_clue = "show stats";
                    else
                        space.TUI_clue = "hide stats";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.show_fixed_bounds = !space.show_fixed_bounds;
                    if(space.show_fixed_bounds)
                        space.TUI_clue = "show fixed bounds";
                    else
                        space.TUI_clue = "hide fixed bounds";
                    break;
                //------------------------------------------------------------
                case '6':
                    space.show_fixed_origin = !space.show_fixed_origin;
                    if(space.show_fixed_origin)
                        space.TUI_clue = "show fixed origin";
                    else
                        space.TUI_clue = "hide fixed origin";
                    break;
                //------------------------------------------------------------
                case '7':
                    space.show_floating_axis = !space.show_floating_axis;
                    if(space.show_floating_axis)
                        space.TUI_clue = "show floating axis";
                    else
                        space.TUI_clue = "hide floating axis";
                    break;
                //------------------------------------------------------------
                case '8':
                    space.show_floating_bounds = !space.show_floating_bounds;
                    if(space.show_floating_bounds)
                        space.TUI_clue = "show floating bounds";
                    else
                        space.TUI_clue = "hide floating bounds";
                    break;
                //------------------------------------------------------------
                case '9':
                    space.show_TUI_clue = !space.show_TUI_clue;
                    if(space.show_TUI_clue)
                        space.TUI_clue = "show UI hint";
                    else
                        space.TUI_clue = "hide UI hint";
                    break;
                //------------------------------------------------------------
                case '0':
                    space.ui_visuals_flip_flop = !space.ui_visuals_flip_flop;
                    space.show_vertices        =  space.ui_visuals_flip_flop;
                    space.show_blanking        =  space.ui_visuals_flip_flop;
                    space.show_palettes        =  space.ui_visuals_flip_flop;
                    space.show_stats           =  space.ui_visuals_flip_flop;
                    space.show_fixed_bounds    =  space.ui_visuals_flip_flop;
                    space.show_fixed_origin    =  space.ui_visuals_flip_flop;
                    space.show_floating_axis   =  space.ui_visuals_flip_flop;
                    space.show_floating_bounds =  space.ui_visuals_flip_flop;
                    space.show_TUI_clue      =  space.ui_visuals_flip_flop;
                    if(space.ui_visuals_flip_flop)
                        space.TUI_clue = "show visuals";
                    else
                        space.TUI_clue = "hide visuals";
                    break;
                //------------------------------------------------------------
                //------------------------------------------------------------
                case ';': // toggle bitmap background
                    space.show_background_bitmap = !space.show_background_bitmap;
                    break;
            } // end switch(event.key.keysym.unicode)
        } // end if(event.type == SDL_KEYDOWN)
    } // end while(in_menu)
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::switch_settings_menu()
{
    bool in_menu = true;

    space.TUI_clue = "switch settings";

    while(in_menu)
    {
        space.render_space();
        space.overlay_switch_settings_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                default:
                    space.TUI_clue = "bad key";
                    break;
                //------------------------------------------------------------
                case SDLK_ESCAPE:
                    in_menu = false;
                    space.TUI_clue = "exit switch settings";
                    break;
                //------------------------------------------------------------
                case '1':
                    space.loop_animation = !space.loop_animation;
                    if(space.loop_animation)
                        space.TUI_clue = "loop animation";
                    else
                        space.TUI_clue = "stop animatiion at ends";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.approximate_frame_rate = !space.approximate_frame_rate;
                    if(space.approximate_frame_rate)
                        space.TUI_clue = "aproximate frame rate";
                    else
                        space.TUI_clue = "as fast as possible";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.show_effects_generation = !space.show_effects_generation;
                    if(space.show_effects_generation)
                        space.TUI_clue = "show live rendering";
                    else
                        space.TUI_clue = "hide live rendering";
                    break;
                //------------------------------------------------------------
                //------------------------------------------------------------
                case '4':
                    space.auto_scale_real = !space.auto_scale_real;
                    if(space.auto_scale_real)
                        space.TUI_clue = "scale dxf import";
                    else
                        space.TUI_clue = "no scale dxf import";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.maintain_real_origin = !space.maintain_real_origin;
                    if(space.maintain_real_origin)
                        space.TUI_clue = "maintain dxf origin";
                    else
                        space.TUI_clue = "ignore dxf origin";
                    break;
                //------------------------------------------------------------
                //------------------------------------------------------------
                case '6':
                    space.no_equivalent_vectors = !space.no_equivalent_vectors;
                    if(space.no_equivalent_vectors)
                        space.TUI_clue = "+ equal vector remove";
                    else
                        space.TUI_clue = "- equal vector remove";
                    break;
                //------------------------------------------------------------
                case '7':
                    space.black_dwell_vertices = !space.black_dwell_vertices;
                    if(space.black_dwell_vertices)
                        space.TUI_clue = "black dwell";
                    else
                        space.TUI_clue = "lit dwell";
                    break;
                //------------------------------------------------------------
                //------------------------------------------------------------
                case '8':
                    space.no_black_match = !space.no_black_match;
                    if(space.no_black_match)
                        space.TUI_clue = "no black match";
                    else
                        space.TUI_clue = "black match palette";
                    break;
                //------------------------------------------------------------
                case '9':
                    space.allow_lit_black = !space.allow_lit_black;
                    if(space.allow_lit_black)
                        space.TUI_clue = "allow lit black";
                    else
                        space.TUI_clue = "blank black";
                    break;
                //------------------------------------------------------------
                //------------------------------------------------------------
                case '0':
                    space.auto_optimize = !space.auto_optimize;
                    if(space.auto_optimize)
                        space.TUI_clue = "optimize file output";
                    else
                        space.TUI_clue = "raw file output";
                    break;
                //------------------------------------------------------------
                case 'a':
                    space.dumb_file_format = !space.dumb_file_format;
                    if(space.dumb_file_format)
                        space.TUI_clue = "ild 4,5";
                    else
                        space.TUI_clue = "ild 0,1,2,(!3)";
                    break;
                //------------------------------------------------------------
                case 'b':
                    space.save_true_color_dxf = !space.save_true_color_dxf;
                    if(space.save_true_color_dxf)
                        space.TUI_clue = "dxf true color";
                    else
                        space.TUI_clue = "dxf palette";
                    break;
                //------------------------------------------------------------
                case 'c':
                    space.auto_flatten_z = !space.auto_flatten_z;
                    if(space.auto_flatten_z)
                        space.TUI_clue = "flatten Z for wave";
                    else
                        space.TUI_clue = "keep Z for wave";
                    break;
                //------------------------------------------------------------
                case 'd':
                    space.invert_wave_output = !space.invert_wave_output;
                    if(space.invert_wave_output)
                        space.TUI_clue = "inverted wave output";
                    else
                        space.TUI_clue = "positive wave output";
                    break;
                //------------------------------------------------------------
                case 'e':
                    space.save_txt_with_color = !space.save_txt_with_color;
                    if(space.save_txt_with_color)
                        space.TUI_clue = "save txt rgb";
                    else
                        space.TUI_clue = "save txt palette & tables";
                    break;
                //------------------------------------------------------------
                case 'f':
                    space.save_txt_unit = !space.save_txt_unit;
                    if(space.save_txt_unit)
                        space.TUI_clue = "save txt unit";
                    else
                        space.TUI_clue = "save txt short";
                    break;
                //------------------------------------------------------------
                case 'g':
                    space.save_txt_color_hex = !space.save_txt_color_hex;
                    if(space.save_txt_color_hex)
                        space.TUI_clue = "save txt hex";
                    else
                        space.TUI_clue = "save txt r g b";
                    break;
                //------------------------------------------------------------
                case 'h':
                    space.save_txt_named_palettes = !space.save_txt_named_palettes;
                    if(space.save_txt_named_palettes)
                        space.TUI_clue = "save named palettes";
                    else
                        space.TUI_clue = "assume named palettes";
                    break;
                //------------------------------------------------------------
                //------------------------------------------------------------
                case ';': // toggle bitmap background
                    space.show_background_bitmap = !space.show_background_bitmap;
                    break;
            } // end switch(event.key.keysym.unicode)
        } // end if(event.type == SDL_KEYDOWN)
    } // end while(in_menu)
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::apply_view_menu()
{
    bool in_menu = true;

    space.simplify_view_angle();

    space.TUI_clue = "view settings";

    while(in_menu)
    {
        space.render_space();
        space.overlay_veiw_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    in_menu = false;
                    space.TUI_clue = "exit view";
                    break;
                //------------------------------------------------------------
                default:
                    check_for_SDL_arrow_keys(event.key.keysym.sym);
                    break;
                //------------------------------------------------------------
                case '1':
                    space.view_angle.x += space.rotation_step;
                    space.TUI_clue = "+ orbit X";
                    break;
                //------------------------------------------------------------
                case '!':
                    space.view_angle.x -= space.rotation_step;
                    space.TUI_clue = "- orbit X";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.view_angle.y += space.rotation_step;
                    space.TUI_clue = "+ orbit Y";
                    break;
                //------------------------------------------------------------
                case '@':
                    space.view_angle.y -= space.rotation_step;
                    space.TUI_clue = "- orbit Y";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.view_angle.z += space.rotation_step;
                    space.TUI_clue = "+ orbit Z";
                    break;
                //------------------------------------------------------------
                case '#':
                    space.view_angle.z -= space.rotation_step;
                    space.TUI_clue = "- orbit Z";
                    break;
                //------------------------------------------------------------
                case '4':
                    space.view_angle += space.rotation_step;
                    space.TUI_clue = "+ orbit XYZ";
                    break;
                //------------------------------------------------------------
                case '$':
                    space.view_angle -= space.rotation_step;
                    space.TUI_clue = "- orbit XYZ";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.view_angle   = 0;
                    space.TUI_clue = "front";
                    break;
                //------------------------------------------------------------
                case '%':
                    space.view_angle.x = 0;
                    space.view_angle.y = pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "back";
                    break;
                //------------------------------------------------------------
                case '6':
                    space.view_angle.x = 0;
                    space.view_angle.y = half_pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "left side";
                    break;
                //------------------------------------------------------------
                case '^':
                    space.view_angle.x = 0;
                    space.view_angle.y = -half_pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "right side";
                    break;
                //------------------------------------------------------------
                case '7':
                    space.view_angle.x = half_pi;
                    space.view_angle.y = 0;
                    space.view_angle.z = 0;
                    space.TUI_clue = "top";
                    break;
                //------------------------------------------------------------
                case '&':
                    space.view_angle.x = -half_pi;
                    space.view_angle.y = 0;
                    space.view_angle.z = 0;
                    space.TUI_clue = "bottom";
                    break;
                //------------------------------------------------------------
                case '8':
                    space.view_angle += space.rotation_step;
                    if(    space.loop_animation
                        || (space.frame_index < space.number_of_frames() - 1)
                      )
                        ++space;
                    space.TUI_clue = "+ frame & orbit";
                    break;
                //------------------------------------------------------------
                case '*':
                    space.view_angle -= space.rotation_step;
                    if(    space.loop_animation
                        || (space.frame_index > 0)
                      )
                        --space;
                    space.TUI_clue = "- frame & orbit";
                    break;
                //------------------------------------------------------------
                case '9': // go to first frame
                    space.frame(0);
                    space.TUI_clue = "first frame";
                    break;
                //------------------------------------------------------------
                case '0': // go to last frame
                    space.frame(space.number_of_frames() - 1);
                    space.TUI_clue = "last frame";
                    break;
                //------------------------------------------------------------
                case 'q':
                    space.view_offset.x += space.displacement_step;
                    space.TUI_clue = "+ pan X";
                    break;
                //------------------------------------------------------------
                case 'Q':
                    space.view_offset.x -= space.displacement_step;
                    space.TUI_clue = "- pan X";
                    break;
                //------------------------------------------------------------
                case 'w':
                    space.view_offset.y += space.displacement_step;
                    space.TUI_clue = "+ pan Y";
                    break;
                //------------------------------------------------------------
                case 'W':
                    space.view_offset.y -= space.displacement_step;
                    space.TUI_clue = "- pan Y";
                    break;
                //------------------------------------------------------------
                case 'e':
                    space.view_offset += space.displacement_step;
                    space.TUI_clue = "+ pan XY";
                    break;
                //------------------------------------------------------------
                case 'E':
                    space.view_offset -= space.displacement_step;
                    space.TUI_clue = "- pan XY";
                    break;
                //------------------------------------------------------------
                case 'r':
                case 'R':
                    space.view_offset = 0.0;
                    space.TUI_clue = "on center";
                    break;
                //------------------------------------------------------------
                case 'a':
                    space.view_scale.x *= space.magnitude_step_up;
                    space.TUI_clue = "+ zoom X";
                    break;
                //------------------------------------------------------------
                case 'A':
                    space.view_scale.x *= space.magnitude_step_dn;
                    space.TUI_clue = "- zoom X";
                    break;
                //------------------------------------------------------------
                case 's':
                    space.view_scale.y *= space.magnitude_step_up;
                    space.TUI_clue = "+ zoom Y";
                    break;
                //------------------------------------------------------------
                case 'S':
                    space.view_scale.y *= space.magnitude_step_dn;
                    space.TUI_clue = "- zoom Y";
                    break;
                //------------------------------------------------------------
                case 'd':
                    space.view_scale *= space.magnitude_step_up;
                    space.TUI_clue = "+ zoom XY";
                    break;
                //------------------------------------------------------------
                case 'D':
                    space.view_scale *= space.magnitude_step_dn;
                    space.TUI_clue = "- zoom XY";
                    break;
                //------------------------------------------------------------
                case 'f':
                case 'F':
                    space.view_scale = 1.0;
                    space.TUI_clue = "no zoom";
                    break;
                //------------------------------------------------------------
                case 'o':
                    space.current_frame().rotate_around_origin();
                    space.view_angle = 0.0;
                    space.TUI_clue = "set rotate frame";
                    break;
                //------------------------------------------------------------
                case 'O':
                    space.TUI_clue = "rotating all";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.rotate_around_origin();
                    space.view_angle = 0.0;
                    space.TUI_clue = "set rotate all";
                    break;
                //------------------------------------------------------------
                case 't':
                    space.current_frame().move();
                    space.view_offset = 0.0;
                    space.TUI_clue = "set offset frame";
                    break;
                //------------------------------------------------------------
                case 'T':
                    space.TUI_clue = "offsetting all";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.move();
                    space.view_offset = 0.0;
                    space.TUI_clue = "set offset all";
                    break;
                //------------------------------------------------------------
                case 'g':
                    space.current_frame().scale_around_origin();
                    space.view_scale = 1.0;
                    space.TUI_clue = "set scale frame";
                    break;
                //------------------------------------------------------------
                case 'G':
                    space.TUI_clue = "scaling all";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.scale_around_origin();
                    space.view_scale = 1.0;
                    space.TUI_clue = "set scale all";
                    break;
                //------------------------------------------------------------
                case 'h':
                    space.current_frame().scale_around_origin();
                    space.current_frame().rotate_around_origin();
                    space.current_frame().move();
                    space.view_angle  = 0.0;
                    space.view_scale  = 1.0;
                    space.view_offset = 0.0;
                    space.TUI_clue = "set view frame";
                    break;
                //------------------------------------------------------------
                case 'H':
                    space.TUI_clue = "setting view all";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.scale_around_origin();
                    space.rotate_around_origin();
                    space.move();
                    space.view_angle  = 0.0;
                    space.view_scale  = 1.0;
                    space.view_offset = 0.0;
                    space.TUI_clue = "set view all";
                    break;
                //------------------------------------------------------------
                case ';': // toggle bitmap background
                    space.show_background_bitmap = !space.show_background_bitmap;
                    break;
            } // end switch(event.key.keysym.unicode)
        } // end if(event.type == SDL_KEYDOWN)
    } // end while(in_menu)
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::palette_transforms_menu()
{
    bool in_menu       = true,
         show_menu_was = space.show_menu;

    space.show_menu = true;
    space.TUI_clue = "palette transforms";

    while(in_menu)
    {
        space.render_space();
        if(space.show_menu)
            space.overlay_palette_tranz_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    in_menu = false;
                    space.TUI_clue = "exit palette transforms";
                    break;
                //------------------------------------------------------------
                default:
                    check_for_SDL_arrow_keys(event.key.keysym.sym);
                    break;
                //------------------------------------------------------------
                case '1':
                    space.current_frame().rainbow_recolor(0);
                    space.TUI_clue = "span X frame";
                    break;
                //------------------------------------------------------------
                case '!':
                    space.TUI_clue = "span X all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(0);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.current_frame().rainbow_recolor(1);
                    space.TUI_clue = "span Y frame";
                    break;
                //------------------------------------------------------------
                case '@':
                    space.TUI_clue = "span Y all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(1);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.current_frame().rainbow_recolor(2);
                    space.TUI_clue = "span Z frame";
                    break;
                //------------------------------------------------------------
                case '#':
                    space.TUI_clue = "span Z all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(2);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '4':
                    space.current_frame().rainbow_recolor(3);
                    space.TUI_clue = "span radially from origin";
                    break;
                //------------------------------------------------------------
                case '$':
                    space.TUI_clue = "span radially from origin";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(3);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.current_frame().rainbow_recolor(4);
                    space.TUI_clue = "span angular around origin";
                    break;
                //------------------------------------------------------------
                case '%':
                    space.TUI_clue = "span angular around origin";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(4);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '6':
                    space.current_frame().rainbow_recolor(5);
                    space.TUI_clue = "span vertices frame";
                    break;
                //------------------------------------------------------------
                case '^':
                    space.TUI_clue = "span vertices all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(5);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '7':
                    space.current_frame().rainbow_recolor(6);
                    space.TUI_clue = "index vertices frame";
                    break;
                //------------------------------------------------------------
                case '&':
                    space.TUI_clue = "index vertices all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(6);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '8':
                    space.current_frame().rainbow_recolor(7);
                    space.TUI_clue = "span segments frame";
                    break;
                //------------------------------------------------------------
                case '*':
                    space.TUI_clue = "span segments all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(7);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '9':
                    space.current_frame().rainbow_recolor(8);
                    space.TUI_clue = "index segments frame";
                    break;
                //------------------------------------------------------------
                case '(':
                    space.TUI_clue = "index segments all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(8);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case '0':
                    space.current_frame().rainbow_recolor(9);
                    space.TUI_clue = "random vertices frame";
                    break;
                //------------------------------------------------------------
                case ')':
                    space.TUI_clue = "random vertices all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(9);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'a':
                    space.current_frame().rainbow_recolor(10);
                    space.TUI_clue = "random segments frame";
                    break;
                //------------------------------------------------------------
                case 'A':
                    space.TUI_clue = "random segments all";
                    space.overlay_please_wait();
                    display_space();
                    space.rainbow_recolor(10);
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'b':
                    space.current_frame().accept_target_palette();
                    space.TUI_clue = "by index frame";
                    break;
                //------------------------------------------------------------
                case 'B':
                    space.TUI_clue = "by index all";
                    space.overlay_please_wait();
                    display_space();
                    space.accept_target_palette();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'c':
                    space.current_frame().best_match_target_palette();
                    space.TUI_clue = "best match frame";
                    break;
                //------------------------------------------------------------
                case 'C':
                    space.TUI_clue = "best match all";
                    space.overlay_please_wait();
                    display_space();
                    space.best_match_target_palette();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'd':
                    space.current_frame().strip_color();
                    space.TUI_clue = "strip color frame";
                    break;
                //------------------------------------------------------------
                case 'D':
                    space.TUI_clue = "strip color all";
                    space.overlay_please_wait();
                    display_space();
                    space.strip_color();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'e':
                    space.current_frame().strip_color_or();
                    space.TUI_clue = "strip color frame";
                    break;
                //------------------------------------------------------------
                case 'E':
                    space.TUI_clue = "strip color all";
                    space.overlay_please_wait();
                    display_space();
                    space.strip_color_or();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'f':
                    space.current_frame().strip_color_avg();
                    space.TUI_clue = "strip color frame";
                    break;
                //------------------------------------------------------------
                case 'F':
                    space.TUI_clue = "strip color all";
                    space.overlay_please_wait();
                    display_space();
                    space.strip_color_avg();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'g':
                    space.current_frame().sync_rgb_and_palette();
                    space.TUI_clue = "24bit to palette frame";
                    break;
                //------------------------------------------------------------
                case 'G':
                    space.TUI_clue = "24bit to palette all";
                    space.overlay_please_wait();
                    display_space();
                    space.sync_rgb_and_palette();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'h':
                    space.current_frame().bit_reduce_to_palette();
                    space.TUI_clue = "888 to 332 frame";
                    break;
                //------------------------------------------------------------
                case 'H':
                    space.TUI_clue = "888 to 332 all";
                    space.overlay_please_wait();
                    display_space();
                    space.bit_reduce_to_palette();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'i':
                    space.current_frame().best_reduce_to_palette();
                    space.TUI_clue = "best reduce to minimum palette";
                    break;
                //------------------------------------------------------------
                case 'I':
                    space.TUI_clue = "best reduce to minimum palette";
                    space.overlay_please_wait();
                    display_space();
                    space.best_reduce_to_palette();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'j':
                    space.current_frame().promote_to_true_color();
                    space.TUI_clue = "promote to 24bit frame";
                    break;
                //------------------------------------------------------------
                case 'J':
                    space.TUI_clue = "promote to 24bit all";
                    space.overlay_please_wait();
                    display_space();
                    space.promote_to_true_color();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'k':
                case 'K':
                    space.TUI_clue = "best reduce all to 1";
                    space.overlay_please_wait();
                    display_space();
                    space.best_reduce_all_to_palette();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case 'l':
                case 'L':
                    space.TUI_clue = "clear unused palettes";
                    space.overlay_please_wait();
                    display_space();
                    space.clear_unused_palettes();
                    space.TUI_clue = "done";
                    break;
                //------------------------------------------------------------
                case ';': // toggle bitmap background
                    space.show_background_bitmap = !space.show_background_bitmap;
                    break;
                //------------------------------------------------------------
                case SDLK_RETURN:
                    space.show_menu = !space.show_menu;
            }
        }
    }
    space.show_menu = show_menu_was;
    return;
}
//############################################################################
void LaserBoy_SDL_GUI::frame_transforms_menu()
{
    bool in_menu       = true,
         show_menu_was = space.show_menu;

    space.show_menu    = true;

    space.TUI_clue = "frame transforms";

    while(in_menu)
    {
        space.render_space();
        if(space.show_frame_stats)
            space.overlay_frame_stats();
        if(space.show_frame_set_stats)
            space.overlay_frame_set_stats();
        if(space.show_menu)
            space.overlay_frame_tranz_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    in_menu = false;
                    space.TUI_clue = "exit frame transforms";
                    break;
                //------------------------------------------------------------
                default:
                    check_for_SDL_arrow_keys(event.key.keysym.sym);
                    break;
                //------------------------------------------------------------
                case '1':
                    space.current_frame().flatten_z();
                    space.TUI_clue = "flatten Z";
                    break;
                //------------------------------------------------------------
                case '!':
                    space.TUI_clue = "flattening Z";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.flatten_z();
                    space.TUI_clue = "flatten Z all";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.current_frame().z_order_vertices(64000);
                    space.TUI_clue = "z order vertices";
                    break;
                //------------------------------------------------------------
                case '@':
                    space.TUI_clue = "Z ordering vertices";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.z_order_vertices(64000);
                    space.TUI_clue = "Z order vertices all";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.current_frame().reverse();
                    space.TUI_clue = "reverse vector order";
                    break;
                //------------------------------------------------------------
                case '#':
                    space.TUI_clue = "reversing vector order";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.frame_reverse();
                    space.TUI_clue = "reverse vector order all";
                    break;
                //------------------------------------------------------------
                case '4':
                    space.current_frame().flip(0);
                    space.TUI_clue = "mirror";
                    break;
                //------------------------------------------------------------
                case '$':
                    space.TUI_clue = "mirroring";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.flip(0);
                    space.TUI_clue = "mirror all";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.current_frame().flip(1);
                    space.TUI_clue = "flip";
                    break;
                //------------------------------------------------------------
                case '%':
                    space.TUI_clue = "flipping";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.flip(1);
                    space.TUI_clue = "flip all";
                    break;
                //------------------------------------------------------------
                case '6':
                    space.current_frame().quarter_turn(0, 1);
                    space.TUI_clue = "quarter turn CW";
                    break;
                //------------------------------------------------------------
                case '^':
                    space.TUI_clue = "quarter turning CW";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.quarter_turn(0, 1);
                    space.TUI_clue = "quarter turn CW all";
                    break;
                //------------------------------------------------------------
                case '7':
                    space.current_frame().quarter_turn(0, 3);
                    space.TUI_clue = "quarter turn CCW";
                    break;
                //------------------------------------------------------------
                case '&':
                    space.TUI_clue = "quarter turning CCW";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.quarter_turn(0, 3);
                    space.TUI_clue = "quarter turn CCW all";
                    break;
                //------------------------------------------------------------
                case '8':
                    space.current_frame().accentuate_the_positive();
                    space.TUI_clue = "accentuate the positive";
                    break;
                //------------------------------------------------------------
                case '*':
                    space.TUI_clue = "accentuating the positive";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.accentuate_the_positive();
                    space.TUI_clue = "accentuate the positive all";
                    break;
                //------------------------------------------------------------
                case '9':
                    space.current_frame().eliminate_the_negative();
                    space.TUI_clue = "eliminate the negative";
                    break;
                //------------------------------------------------------------
                case '(':
                    space.TUI_clue = "eliminating the negative";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.eliminate_the_negative();
                    space.TUI_clue = "eliminate the negative all";
                    break;
                //------------------------------------------------------------
                case '0':
                    space.current_frame().impose_bit_resolution();
                    space.TUI_clue = "imposed bit resolution";
                    break;
                //------------------------------------------------------------
                case ')':
                    space.TUI_clue = "imposing bit resolution";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.impose_bit_resolution();
                    space.TUI_clue = "imposed bit resolution all";
                    break;
                //------------------------------------------------------------
                case 'a':
                    space.current_frame().convert_black_to_blank();
                    space.TUI_clue = "black to blank";
                    break;
                //------------------------------------------------------------
                case 'A':
                    space.TUI_clue = "converting black to blank";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.convert_black_to_blank();
                    space.TUI_clue = "black to blank all";
                    break;
                //------------------------------------------------------------
                case 'b':
                    space.current_frame().convert_blank_to_black();
                    space.TUI_clue = "blank to black";
                    break;
                //------------------------------------------------------------
                case 'B':
                    space.TUI_clue = "converting black to blank";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.convert_blank_to_black();
                    space.TUI_clue = "blank to black all";
                    break;
                //------------------------------------------------------------
                case 'c':
                    space.current_frame().impose_black_level();
                    space.TUI_clue = "apply black level";
                    break;
                //------------------------------------------------------------
                case 'C':
                    space.TUI_clue = "applying black level";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.impose_black_level();
                    space.TUI_clue = "black leveled all";
                    break;
                //------------------------------------------------------------
                case 'd':
                    space.current_frame().to_dots();
                    space.TUI_clue = "to dots";
                    break;
                //------------------------------------------------------------
                case 'D':
                    space.TUI_clue = "converting to dots";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.to_dots();
                    space.TUI_clue = "to dots all";
                    break;
                //------------------------------------------------------------
                case 'e':
                    space.current_frame().reduce_blank_vectors();
                    space.TUI_clue = "reduced blank span";
                    break;
                //------------------------------------------------------------
                case 'E':
                    space.TUI_clue = "reduce blank vectors";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.reduce_blank_vectors();
                    space.TUI_clue = "reduced blank span all";
                    break;
                //------------------------------------------------------------
                case 'f':
                    space.current_frame().remove_dwell_vertices();
                    space.TUI_clue = "removed angle dwell";
                    break;
                //------------------------------------------------------------
                case 'F':
                    space.TUI_clue = "removing dwell vertices";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.remove_dwell_vertices();
                    space.TUI_clue = "removed angle dwell all";
                    break;
                //------------------------------------------------------------
                case 'g':
                    space.current_frame().remove_dots();
                    space.TUI_clue = "removed dots";
                    break;
                //------------------------------------------------------------
                case 'G':
                    space.TUI_clue = "removing dots";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.remove_dots();
                    space.TUI_clue = "removed dots all";
                    break;
                //------------------------------------------------------------
                case 'h':
                    space.current_frame().reduce_lit_vectors();
                    space.TUI_clue = "reduced vectors";
                    break;
                //------------------------------------------------------------
                case 'H':
                    space.TUI_clue = "reducing vectors";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.reduce_lit_vectors();
                    space.TUI_clue = "reduced vectors all";
                    break;
                //------------------------------------------------------------
                case 'i':
                    space.current_frame().blank_equivalent_vectors();
                    space.TUI_clue = "removed equal vectors";
                    break;
                //------------------------------------------------------------
                case 'I':
                    space.TUI_clue = "removing equal vectors";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.blank_equivalent_vectors();
                    space.TUI_clue = "removed equal vectors";
                    break;
                //------------------------------------------------------------
                case 'j':
                    space.current_frame().bond_segments();
                    space.TUI_clue = "bonded segments";
                    break;
                //------------------------------------------------------------
                case 'J':
                    space.TUI_clue = "bonding segments";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.bond_segments();
                    space.TUI_clue = "bonded segments all";
                    break;
                //------------------------------------------------------------
                case 'k':
                    space.current_frame().minimize();
                    space.TUI_clue = "minimized vectors";
                    break;
                //------------------------------------------------------------
                case 'K':
                    space.TUI_clue = "minimizing vectors";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.minimize();
                    space.TUI_clue = "minimized vectors all";
                    break;
                //------------------------------------------------------------
                case 'l':
                    space.current_frame().fracture_segments();
                    space.TUI_clue = "fractured segments";
                    break;
                //------------------------------------------------------------
                case 'L':
                    space.TUI_clue = "fracturing segments";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.fracture_segments();
                    space.TUI_clue = "fractured segments all";
                    break;
                //------------------------------------------------------------
                case 'm':
                    space.current_frame().reorder_segments();
                    space.TUI_clue = "reordered segments";
                    break;
                //------------------------------------------------------------
                case 'M':
                    space.TUI_clue = "reordering segments";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.reorder_segments();
                    space.TUI_clue = "reordered segments all";
                    break;
                //------------------------------------------------------------
                case 'n':
                    space.current_frame().add_dwell_vertices();
                    space.TUI_clue = "added angle dwell";
                    break;
                //------------------------------------------------------------
                case 'N':
                    space.TUI_clue = "adding angle dwell";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.add_dwell_vertices();
                    space.TUI_clue = "added angle dwell all";
                    break;
                //------------------------------------------------------------
                case 'o':
                    space.current_frame().add_lit_span_vertices();
                    space.TUI_clue = "added lit distance span";
                    break;
                //------------------------------------------------------------
                case 'O':
                    space.TUI_clue = "adding lit distance span";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.add_lit_span_vertices();
                    space.TUI_clue = "added lit distance span all";
                    break;
                //------------------------------------------------------------
                case 'p':
                    space.current_frame().add_blank_span_vertices();
                    space.TUI_clue = "added blank distance span";
                    break;
                //------------------------------------------------------------
                case 'P':
                    space.TUI_clue = "adding blank distance span";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.add_blank_span_vertices();
                    space.TUI_clue = "added blank distance span all";
                    break;
                //------------------------------------------------------------
                case 'q':
                    space.current_frame().enhance_dots();
                    space.TUI_clue = "enhanced dots";
                    break;
                //------------------------------------------------------------
                case 'Q':
                    space.TUI_clue = "enhancing dots";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.enhance_dots();
                    space.TUI_clue = "enhanced dots";
                    break;
                //------------------------------------------------------------
                case 'r':
                    space.current_frame().optimize();
                    space.TUI_clue = "optimized";
                    break;
                //------------------------------------------------------------
                case 'R':
                    space.TUI_clue = "optimizing";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.optimize();
                    space.TUI_clue = "optimized all";
                    break;
                //------------------------------------------------------------
                case ';': // toggle bitmap background
                    space.show_background_bitmap = !space.show_background_bitmap;
                    break;
                //------------------------------------------------------------
                case SDLK_RETURN:
                    space.show_menu = !space.show_menu;
            }
        }
    }
    space.show_menu = show_menu_was;
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::frame_set_transforms_menu()
{
    bool in_menu       = true,
         show_menu_was = space.show_menu;
    int  temp_int;

    space.show_menu    = true;

    space.TUI_clue = "frame set transforms";

    while(in_menu)
    {
        space.render_space();
        if(space.show_menu)
            space.overlay_frame_set_tranz_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    in_menu = false;
                    space.TUI_clue = "action canceled";
                    break;
                //------------------------------------------------------------
                default:
                    check_for_SDL_arrow_keys(event.key.keysym.sym);
                    break;
                //------------------------------------------------------------
                case '1':
                    space.TUI_clue = "deleting selected frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.delete_selected_frames();
                    space.TUI_clue = "deleted selected frames";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.TUI_clue = "trimming to selected frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.trim_to_selected_frames();
                    space.TUI_clue = "trimmed to selected frames";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.TUI_clue = "deleting redundant frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.delete_redundant_frames();
                    space.TUI_clue = "redundant frames deleted";
                    break;
                //------------------------------------------------------------
                case '4':
                    space.TUI_clue = "copying selected frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space = space.selected_frames() + space;
                    space.TUI_clue = "copied frames to begining";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.TUI_clue = "copying selected frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space += space.selected_frames();
                    space.TUI_clue = "copied frames to end";
                    break;
                //------------------------------------------------------------
                case '6':
                    space.TUI_clue = "moving selected frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    {
                        LaserBoy_frame_set temp(space.selected_frames());
                        space.delete_selected_frames();
                        temp += space;
                        space = temp;
                    }
                    space.TUI_clue = "moved frames to begining";
                    break;
                //------------------------------------------------------------
                case '7':
                    space.TUI_clue = "moving selected frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    {
                        LaserBoy_frame_set temp(space.selected_frames());
                        space.delete_selected_frames();
                        space += temp;
                    }
                    space.TUI_clue = "moved frames to end";
                    break;
                //------------------------------------------------------------
                case '8':
                    {
                        LaserBoy_frame_set temp(&space);
                        temp += LaserBoy_frame(&space, space.target_palette_index, 0); // not quite an empty frame!
                        space = temp + space;
                    }
                    space.TUI_clue = "add frame to begining";
                    break;
                //------------------------------------------------------------
                case '9':
                    {
                        int                i;
                        LaserBoy_frame_set temp(&space);
                        for(i = 0; i <= space.frame_index; i++)
                            temp += space[i];
                        temp += LaserBoy_frame(&space, space.target_palette_index, 0);
                        for(i = space.frame_index + 1; i < space.number_of_frames(); i++)
                            temp += space[i];
                        space = temp;
                    }
                    space.TUI_clue = "add after current frame";
                    break;
                //------------------------------------------------------------
                case '0':
                    space += LaserBoy_frame(&space, space.target_palette_index, 0);
                    space.TUI_clue = "add frame to end";
                    break;
                //------------------------------------------------------------
                case 'a':
                    space.TUI_clue = "reversing order of frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.reverse();
                    space.TUI_clue = "reverse order";
                    break;
                //------------------------------------------------------------
                case 'b':
                    space.TUI_clue = "reversing selected frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.reverse_selected();
                    space.TUI_clue = "reverse selected";
                    break;
                //------------------------------------------------------------
                case 'c':
                    space.TUI_clue = "new first frame";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.make_current_first();
                    space.TUI_clue = "new first frame";
                    break;
                //------------------------------------------------------------
                case 'd':
                    space.TUI_clue = "random order frames";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();
                    space.random_order();
                    space.TUI_clue = "random order";
                    break;
                //------------------------------------------------------------
                case 'e':
                    space.TUI_clue = "remove every n-th frames";
                    space.render_space();
                    display_space();
                    temp_int = display_prompt_and_echo_int("enter value of n");
                    if(prompt_escaped)
                        break;
                    space.overlay_please_wait();
                    space.delete_every_nth_frame(temp_int);
                    space.TUI_clue = "n-th frame delete";
                    break;
                //------------------------------------------------------------
                case '>': // explode current frame
                    space.explode_current_frame();
                    space.TUI_clue = "explode segments";
                    break;
                //------------------------------------------------------------
                case '<': // collapse selected frames
                    space.collapse_selected_frames();
                    space.TUI_clue = "collapse segments";
                    break;
                //------------------------------------------------------------
                case '!': // explode all frames
                    space.explode_all_frames();
                    space.TUI_clue = "explode segments";
                    break;
                //------------------------------------------------------------
                case ' ': // [Space] select, deselect current frame
                    space.current_frame().is_selected = !(space.current_frame().is_selected);
                    if(space.current_frame().is_selected)
                        space.TUI_clue = "select frame";
                    else
                        space.TUI_clue = "deselect frame";
                    break;
                //------------------------------------------------------------
                case ';': // toggle bitmap background
                    space.show_background_bitmap = !space.show_background_bitmap;
                    break;
                //------------------------------------------------------------
                case SDLK_RETURN:
                    space.show_menu = !space.show_menu;
            }
        }
    }
    space.show_menu = show_menu_was;
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::draw_color_blank_menu()
{
    bool  in_menu           = true,
          show_vertices_was = space.show_vertices,
          show_blanking_was = space.show_blanking,
          show_menu_was     = space.show_menu;

    int                temp_int,
                       number_of_sides;

    double             temp_double;

    LaserBoy_3D_double temp_3D_float;

    space.show_menu         = true;
    space.show_vertices     = true;
    space.show_blanking     = true;
    space.show_cursors      = true;

    space.TUI_clue = "draw color blank";

    while(in_menu)
    {
        space.render_space();
        if(space.show_menu)
            space.overlay_draw_color_blank_menu();
        if(space.show_frame_stats)
            space.overlay_frame_stats();
        if(space.show_frame_set_stats)
            space.overlay_frame_set_stats();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                //------------------------------------------------------------
                case SDLK_ESCAPE:
                    in_menu   = false;
                    space.show_cursors  = false;
                    space.show_vertices = show_vertices_was;
                    space.show_blanking = show_blanking_was;
                    space.TUI_clue = "exit draw color blank";
                    break;
                //------------------------------------------------------------
                case SDLK_BACKSPACE:
                    space.current_frame().remove_vertex(space.current_frame().spider);
                    space.TUI_clue = "delete vertex";
                    break;
                //------------------------------------------------------------
                default:
                    check_for_SDL_arrow_keys(event.key.keysym.sym);
                    break;
                //------------------------------------------------------------
                case '1':
                    space.view_angle.x += space.rotation_step;
                    space.TUI_clue = "+ orbit X";
                    break;
                //------------------------------------------------------------
                case '!':
                    space.view_angle.x -= space.rotation_step;
                    space.TUI_clue = "- orbit X";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.view_angle.y += space.rotation_step;
                    space.TUI_clue = "+ orbit Y";
                    break;
                //------------------------------------------------------------
                case '@':
                    space.view_angle.y -= space.rotation_step;
                    space.TUI_clue = "- orbit Y";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.view_angle.z += space.rotation_step;
                    space.TUI_clue = "+ orbit Z";
                    break;
                //------------------------------------------------------------
                case '#':
                    space.view_angle.z -= space.rotation_step;
                    space.TUI_clue = "- orbit Z";
                    break;
                //------------------------------------------------------------
                case '4':
                    space.view_angle += space.rotation_step;
                    space.TUI_clue = "+ orbit XYZ";
                    break;
                //------------------------------------------------------------
                case '$':
                    space.view_angle -= space.rotation_step;
                    space.TUI_clue = "- orbit XYZ";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.view_angle   = 0;
                    space.TUI_clue = "front";
                    break;
                //------------------------------------------------------------
                case '%':
                    space.view_angle.x = 0;
                    space.view_angle.y = pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "back";
                    break;
                //------------------------------------------------------------
                case '6':
                    space.view_angle.x = 0;
                    space.view_angle.y = half_pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "left side";
                    break;
                //------------------------------------------------------------
                case '^':
                    space.view_angle.x = 0;
                    space.view_angle.y = -half_pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "right side";
                    break;
                //------------------------------------------------------------
                case '7':
                    space.view_angle.x = half_pi;
                    space.view_angle.y = 0;
                    space.view_angle.z = 0;
                    space.TUI_clue = "top";
                    break;
                //------------------------------------------------------------
                case '&':
                    space.view_angle.x = -half_pi;
                    space.view_angle.y = 0;
                    space.view_angle.z = 0;
                    space.TUI_clue = "bottom";
                    break;
                //------------------------------------------------------------
                case '8':
                    space.view_angle += space.rotation_step;
                    if(    space.loop_animation
                        || (space.frame_index < space.number_of_frames() - 1)
                      )
                        ++space;
                    space.TUI_clue = "+ frame & orbit";
                    break;
                //------------------------------------------------------------
                case '*':
                    space.view_angle -= space.rotation_step;
                    if(    space.loop_animation
                        || (space.frame_index > 0)
                      )
                        --space;
                    space.TUI_clue = "- frame & orbit";
                    break;
                //------------------------------------------------------------
                case '[':
                    space.current_frame().increment_egg();
                    space.TUI_clue = "+ start vertex select";
                    break;
                //------------------------------------------------------------
                case '{':
                    space.current_frame().deccrement_egg();
                    space.TUI_clue = "- start vertex select";
                    break;
                //------------------------------------------------------------
                case ']':
                    space.current_frame().increment_spider();
                    space.TUI_clue = "+ end vertex select";
                    break;
                //------------------------------------------------------------
                case '}':
                    space.current_frame().deccrement_spider();
                    space.TUI_clue = "- end vertex select";
                    break;
                //------------------------------------------------------------
                case'\\':
                    space.current_frame().spider = (space.current_frame().number_of_vertices() - 1);
                    space.TUI_clue = "last vertex select";
                    break;
                //------------------------------------------------------------
                case '|':
                    space.current_frame().egg = 0;
                    space.TUI_clue = "first vertex select";
                    break;
                //------------------------------------------------------------
                case 'u':
                    space.current_frame().next_segment_select();
                    space.TUI_clue = "+ lit segment";
                    break;
                //------------------------------------------------------------
                case 'U':
                    space.current_frame().previous_segment_select();
                    space.TUI_clue = "- lit segment";
                    break;
                //------------------------------------------------------------
                case 'i':
                    space.current_frame().next_segment_egg();
                    space.TUI_clue = "+ start lit segment";
                    break;
                //------------------------------------------------------------
                case 'I':
                    space.current_frame().previous_segment_egg();
                    space.TUI_clue = "- start lit segment";
                    break;
                //------------------------------------------------------------
                case 'o':
                    space.current_frame().next_segment_spider();
                    space.TUI_clue = "+ end lit segment";
                    break;
                //------------------------------------------------------------
                case 'O':
                    space.current_frame().previous_segment_spider();
                    space.TUI_clue = "- end lit segment";
                    break;
                //------------------------------------------------------------
                case '/': // show frame set stats
                    space.show_frame_set_stats = !space.show_frame_set_stats;
                    space.TUI_clue = "frame set stats";
                    break;
                //------------------------------------------------------------
                case '?': // show frame stats
                    space.show_frame_stats = !space.show_frame_stats;
                    space.TUI_clue = "frame stats";
                    break;
                //------------------------------------------------------------
                case 'q':
                    temp_int = display_prompt_and_echo_int("set value of X");
                    if(prompt_escaped)
                        break;
                    space.current_frame().at(space.current_frame().spider).x = temp_int;
                    space.TUI_clue = "X set";
                    break;
                //------------------------------------------------------------
                case 'w':
                    temp_int = display_prompt_and_echo_int("set value of Y");
                    if(prompt_escaped)
                        break;
                    space.current_frame().at(space.current_frame().spider).y = temp_int;
                    space.TUI_clue = "Y set";
                    break;
                //------------------------------------------------------------
                case 'e':
                    temp_int = display_prompt_and_echo_int("set value of Z");
                    if(prompt_escaped)
                        break;
                    space.current_frame().at(space.current_frame().spider).z = temp_int;
                    space.TUI_clue = "Z set";
                    break;
                //------------------------------------------------------------
                case 'p':
                    space.selected_palette_color++;
                    if(space.selected_palette_color >= space.current_palette().number_of_colors())
                        space.selected_palette_color = 0;
                    space.TUI_clue = "+ color index";
                    break;
                //------------------------------------------------------------
                case 'P':
                    space.selected_palette_color--;
                    if(space.selected_palette_color < 0)
                        space.selected_palette_color = space.current_palette().number_of_colors() - 1;
                    space.TUI_clue = "- color index";
                    break;
                //------------------------------------------------------------
                case 'a':
                    space.current_frame().unblank_all_vertices();
                    space.TUI_clue = "unblank all vectors";
                    break;
                //------------------------------------------------------------
                case 'A':
                    space.unblank_all_vertices();
                    space.TUI_clue = "unblank all frames";
                    break;
                //------------------------------------------------------------
                case 'b':
                    space.current_frame().blank_vertices();
                    space.TUI_clue = "blank selected vectors";
                    break;
                //------------------------------------------------------------
                case 'B':
                    space.current_frame().unblank_vertices();
                    space.TUI_clue = "unblank selected vectors";
                    break;
                //------------------------------------------------------------
                case 'n':
                    space.current_frame().black_vertices();
                    space.TUI_clue = "bond black selected vectors";
                    break;
                //------------------------------------------------------------
                case 'N':
                    space.current_frame().unblack_vertices();
                    space.TUI_clue = "unbond unblack selected vectors";
                    break;
                //------------------------------------------------------------
                case 'c':
                    space.current_frame().recolor_select();
                    space.TUI_clue = "color selected vectors";
                    break;
                //------------------------------------------------------------
                case 'C':
                    space.selected_palette_color = space.current_frame().at(space.current_frame().spider).c;
                    space.TUI_clue = "get color";
                    break;
                //------------------------------------------------------------
                case 'h':
                    space.current_frame().recolor_black_in_select();
                    space.TUI_clue = "recolor black selected vectors";
                    break;
                //------------------------------------------------------------
                case '.':
                    space.current_frame().add_new_vertex(space.current_frame().spider);
                    if(space.current_frame().spider < (space.current_frame().number_of_vertices() - 1))
                        space.TUI_clue = "split vector";
                    else
                        space.TUI_clue = "add vector";
                    break;
                //------------------------------------------------------------
                case ',':
                    space.current_frame().reorder_from_egg();
                    space.TUI_clue = "reorder from egg";
                    break;
                //------------------------------------------------------------
                case '<':
                    space.current_frame().reorder_from_spider();
                    space.TUI_clue = "reorder from spider";
                    break;
                //------------------------------------------------------------
                case '\'':
                    space.current_frame().selected_to_dots();
                    space.TUI_clue = "to dots";
                    break;
                //------------------------------------------------------------
                case 't':
                    draw_text_menu();
                    break;
                //------------------------------------------------------------
                case 'k':
                    {
                        int number_of_sides = display_prompt_and_echo_int("enter number of sides");
                        if(prompt_escaped)
                            break;
                        space.current_frame()
                            += LaserBoy_segment(&space,
                                                space.current_frame().at(space.current_frame().egg),
                                                space.current_frame().at(space.current_frame().spider),
                                                number_of_sides
                                               );
                    }
                    break;
                //------------------------------------------------------------
                case 'd':
                    draw_circular_arc_menu();
                    break;
                //------------------------------------------------------------
                case 'D':
                    draw_elliptical_arc_menu();
                    break;
                //------------------------------------------------------------
                case 'l':
                    space.current_frame().connect_the_dots();
                    space.TUI_clue = "draw line";
                    break;
                //------------------------------------------------------------
                case 's':
                    space.current_frame().break_segment();
                    space.TUI_clue = "break segment";
                    break;
                //------------------------------------------------------------
                case 'x':
                    temp_double = space.current_frame().at(space.current_frame().spider).x + space.displacement_step;
                    if(temp_double <= LASERBOY_MAX_SHORT)
                        space.current_frame().at(space.current_frame().spider).x = (u_short)temp_double;
                    space.TUI_clue = "+ X";
                    break;
                //------------------------------------------------------------
                case 'X':
                    temp_double = space.current_frame().at(space.current_frame().spider).x - space.displacement_step;
                    if(temp_double >= LASERBOY_MIN_SHORT)
                        space.current_frame().at(space.current_frame().spider).x = (u_short)temp_double;
                    space.TUI_clue = "- X";
                    break;
                //------------------------------------------------------------
                case 'y':
                    temp_double = space.current_frame().at(space.current_frame().spider).y + space.displacement_step;
                    if(temp_double <= LASERBOY_MAX_SHORT)
                        space.current_frame().at(space.current_frame().spider).y = (u_short)temp_double;
                    space.TUI_clue = "+ Y";
                    break;
                //------------------------------------------------------------
                case 'Y':
                    temp_double = space.current_frame().at(space.current_frame().spider).y - space.displacement_step;
                    if(temp_double >= LASERBOY_MIN_SHORT)
                        space.current_frame().at(space.current_frame().spider).y = (u_short)temp_double;
                    space.TUI_clue = "- Y";
                    break;
                //------------------------------------------------------------
                case 'z':
                    temp_double = space.current_frame().at(space.current_frame().spider).z + space.displacement_step;
                    if(temp_double <= LASERBOY_MAX_SHORT)
                        space.current_frame().at(space.current_frame().spider).z = (u_short)temp_double;
                    space.TUI_clue = "+ Z";
                    break;
                //------------------------------------------------------------
                case 'Z':
                    temp_double = space.current_frame().at(space.current_frame().spider).z - space.displacement_step;
                    if(temp_double >= LASERBOY_MIN_SHORT)
                        space.current_frame().at(space.current_frame().spider).z = (u_short)temp_double;
                    space.TUI_clue = "- Z";
                    break;
                //------------------------------------------------------------
                case 'f':
                    if(space.current_frame().spider) // is not the origin
                    {
                       temp_3D_float
                        = rotate_vertex_on_coordinates_z(   space.current_frame().at(space.current_frame().spider)
                                                       , space.current_frame().at(space.current_frame().spider - 1)
                                                       , space.rotation_step
                                                     );
                        if(!LaserBoy_bounds_check(temp_3D_float, LASERBOY_CUBE))
                            space.current_frame().at(space.current_frame().spider) = temp_3D_float;
                        space.TUI_clue = "+ rotate on Z axis";
                    }
                    else
                        space.TUI_clue = "not at zero vertex";
                    break;
                //------------------------------------------------------------
                case 'F':
                    if(space.current_frame().spider) // is not the origin
                    {

                        temp_3D_float
                        = rotate_vertex_on_coordinates_z(   space.current_frame().at(space.current_frame().spider)
                                                       , space.current_frame().at(space.current_frame().spider - 1)
                                                       , -space.rotation_step
                                                     );
                        if(!LaserBoy_bounds_check(temp_3D_float, LASERBOY_CUBE))
                            space.current_frame().at(space.current_frame().spider) = temp_3D_float;
                        space.TUI_clue = "- rotate on Z axis";
                    }
                    else
                        space.TUI_clue = "not at zero vertex";
                    break;
                //------------------------------------------------------------
                case 'g':
                    if(space.current_frame().spider) // is not the origin
                    {
                        temp_3D_float
                        = rotate_vertex_on_coordinates_y(   space.current_frame().at(space.current_frame().spider)
                                                       , space.current_frame().at(space.current_frame().spider - 1)
                                                       , space.rotation_step
                                                     );
                        if(!LaserBoy_bounds_check(temp_3D_float, LASERBOY_CUBE))
                            space.current_frame().at(space.current_frame().spider) = temp_3D_float;
                        space.TUI_clue = "+ rotate on Y axis";
                    }
                    else
                        space.TUI_clue = "not at zero vertex";
                    break;
                //------------------------------------------------------------
                case 'G':
                    if(space.current_frame().spider) // is not the origin
                    {
                        temp_3D_float
                        = rotate_vertex_on_coordinates_y(   space.current_frame().at(space.current_frame().spider)
                                                       , space.current_frame().at(space.current_frame().spider - 1)
                                                       , -space.rotation_step
                                                     );
                        if(!LaserBoy_bounds_check(temp_3D_float, LASERBOY_CUBE))
                            space.current_frame().at(space.current_frame().spider) = temp_3D_float;
                        space.TUI_clue = "- rotate on Y axis";
                    }
                    else
                        space.TUI_clue = "not at zero vertex";
                    break;
                //------------------------------------------------------------
                case 'm':
                    if(space.current_frame().spider) // is not the origin
                    {
                    temp_3D_float
                    = scale_vertex_on_coordinates(   space.current_frame().at(space.current_frame().spider)
                                                , space.current_frame().at(space.current_frame().spider - 1)
                                                , LaserBoy_3D_double(space.magnitude_step_up, space.magnitude_step_up, space.magnitude_step_up)
                                              );
                    if(!LaserBoy_bounds_check(temp_3D_float, LASERBOY_CUBE))
                        space.current_frame().at(space.current_frame().spider) = temp_3D_float;
                        space.TUI_clue = "+ magnitude of vector";
                    }
                    else
                        space.TUI_clue = "not at zero vertex";
                    break;
                //------------------------------------------------------------
                case 'M':
                    if(space.current_frame().spider) // is not the origin
                    {
                        temp_3D_float
                        = scale_vertex_on_coordinates(   space.current_frame().at(space.current_frame().spider)
                                                    , space.current_frame().at(space.current_frame().spider - 1)
                                                    , LaserBoy_3D_double(space.magnitude_step_dn, space.magnitude_step_dn, space.magnitude_step_dn)
                                                  );
                        if(!LaserBoy_bounds_check(temp_3D_float, LASERBOY_CUBE))
                            space.current_frame().at(space.current_frame().spider) = temp_3D_float;
                        space.TUI_clue = "- magnitude of vector";
                    }
                    else
                        space.TUI_clue = "not at zero vertex";
                    break;
                //------------------------------------------------------------
                case ';': // toggle bitmap background
                    space.show_background_bitmap = !space.show_background_bitmap;
                    break;
                //------------------------------------------------------------
                case SDLK_RETURN:
                    space.show_menu = !space.show_menu;
            } // end switch(event.key.keysym.unicode)
        } // end else if(event.type == SDL_KEYDOWN)
    } // end while(in_menu)
    space.show_menu = show_menu_was;
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::draw_text_menu()
{
    double x_rotation;
    string font,
           text;
    //------------------------------------------------------------------------
    font = display_prompt_and_echo_name("enter font (ild) file name");
    if(prompt_escaped)
        return;
    space.render_space();
    if(space.show_menu)
        space.overlay_draw_color_blank_menu();
    if(space.show_frame_stats)
        space.overlay_frame_stats();
    if(space.show_frame_set_stats)
        space.overlay_frame_set_stats();
    display_space();
    //------------------------------------------------------------------------
    if(font.size() == 0)
        font = "lucida.ild";
    else if(font.size() < 5 || font.substr(font.size() - 4) != ".ild")
        font += ".ild";
    //------------------------------------------------------------------------
    x_rotation = display_prompt_and_echo_double("enter angle of rotation around vector");
    if(prompt_escaped)
        return;
    space.render_space();
    if(space.show_menu)
        space.overlay_draw_color_blank_menu();
    if(space.show_frame_stats)
        space.overlay_frame_stats();
    if(space.show_frame_set_stats)
        space.overlay_frame_set_stats();
    display_space();
    //------------------------------------------------------------------------
    text = display_prompt_and_echo_string("text to render                       ");
    if(prompt_escaped)
        return;
    //------------------------------------------------------------------------
    space.current_frame()
    += LaserBoy_segment(&space,
                        space.current_frame().at(space.current_frame().egg),
                        space.current_frame().at(space.current_frame().spider),
                        x_rotation,
                        LASERBOY_ILD_SHARE + font,
                        text
                       );
    //------------------------------------------------------------------------
    space.current_frame().at(space.current_frame().egg).blank();
    //------------------------------------------------------------------------
    space.TUI_clue = "text rendered";
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::draw_circular_arc_menu()
{
    double  arc_angle,
            radii_ratio;

    arc_angle = display_prompt_and_echo_double("enter circular arc angle in degrees, (0 = 360)");
    if(prompt_escaped)
        return;

    space.render_space();
    if(space.show_menu)
        space.overlay_draw_color_blank_menu();
    display_space();

    if(space.current_frame().number_of_vertices() > 1)
    {
        space.current_frame()
        += LaserBoy_segment(&space,
                            space.current_frame().at(space.current_frame().egg),
                            space.current_frame().at(space.current_frame().spider),
                            arc_angle
                           );
        space.TUI_clue = "circular arc";
    }
    else
        space.TUI_clue = "All Your Base Are Belong To Us!";
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::draw_elliptical_arc_menu()
{
    double  arc_angle,
            radii_ratio;

    arc_angle = display_prompt_and_echo_double("enter elliptical arc angle in degrees, (0 = 360)");

    if(prompt_escaped)
        return;

    space.render_space();
    if(space.show_menu)
        space.overlay_draw_color_blank_menu();
    display_space();

    radii_ratio = display_prompt_and_echo_double("enter minor / major radii ratio");
    if(prompt_escaped)
        return;

    space.render_space();
    if(space.show_menu)
        space.overlay_draw_color_blank_menu();
    display_space();

    if(space.current_frame().number_of_vertices() > 1)
    {
        space.current_frame()
            += LaserBoy_segment(&space,
                                space.current_frame().at(space.current_frame().egg),
                                space.current_frame().at(space.current_frame().spider),
                                arc_angle,
                                radii_ratio
                               );
        space.TUI_clue = "elliptical arc";
    }
    else
        space.TUI_clue = "All Your Base Are Belong To Us!";
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::move_scale_rotate_menu()
{
    bool    in_menu           = true,
            show_vertices_was = space.show_vertices,
            show_blanking_was = space.show_blanking,
            show_menu_was     = space.show_menu;

    char formatted_string[80] = {0};

    u_short offset;

    double  temp_double;

    space.show_menu     = true;
    space.show_vertices = true;
    space.show_blanking = true;
    space.show_cursors  = true;
    space.show_fulcrum  = true;

    space.TUI_clue = "move scale rotate";

    while(in_menu)
    {
        space.render_space();
        if(space.show_menu)
            space.overlay_move_scale_rotate_menu();
        if(space.show_frame_stats)
            space.overlay_frame_stats();
        if(space.show_frame_set_stats)
            space.overlay_frame_set_stats();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    in_menu         = false;
                    space.show_vertices = show_vertices_was;
                    space.show_blanking = show_blanking_was;
                    space.show_cursors  = false;
                    space.show_fulcrum  = false;
                    space.TUI_clue = "exit move scale rotate";
                    break;
                //------------------------------------------------------------
                default:
                    check_for_SDL_arrow_keys(event.key.keysym.sym);
                    break;
                //------------------------------------------------------------
                case '1':
                    space.view_angle.x += space.rotation_step;
                    space.TUI_clue = "+ orbit X";
                    break;
                //------------------------------------------------------------
                case '!':
                    space.view_angle.x -= space.rotation_step;
                    space.TUI_clue = "- orbit X";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.view_angle.y += space.rotation_step;
                    space.TUI_clue = "+ orbit Y";
                    break;
                case '@':
                    space.view_angle.y -= space.rotation_step;
                    space.TUI_clue = "- orbit Y";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.view_angle.z += space.rotation_step;
                    space.TUI_clue = "+ orbit Z";
                    break;
                //------------------------------------------------------------
                case '#':
                    space.view_angle.z -= space.rotation_step;
                    space.TUI_clue = "- orbit Z";
                    break;
                //------------------------------------------------------------
                case '4':
                    space.view_angle += space.rotation_step;
                    space.TUI_clue = "+ orbit XYZ";
                    break;
                //------------------------------------------------------------
                case '$':
                    space.view_angle -= space.rotation_step;
                    space.TUI_clue = "- orbit XYZ";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.view_angle   = 0;
                    space.TUI_clue = "front";
                    break;
                //------------------------------------------------------------
                case '%':
                    space.view_angle.x = 0;
                    space.view_angle.y = pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "back";
                    break;
                //------------------------------------------------------------
                case '6':
                    space.view_angle.x = 0;
                    space.view_angle.y = half_pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "left side";
                    break;
                //------------------------------------------------------------
                case '^':
                    space.view_angle.x = 0;
                    space.view_angle.y = -half_pi;
                    space.view_angle.z = 0;
                    space.TUI_clue = "right side";
                    break;
                //------------------------------------------------------------
                case '7':
                    space.view_angle.x = half_pi;
                    space.view_angle.y = 0;
                    space.view_angle.z = 0;
                    space.TUI_clue = "top";
                    break;
                //------------------------------------------------------------
                case '&':
                    space.view_angle.x = -half_pi;
                    space.view_angle.y = 0;
                    space.view_angle.z = 0;
                    space.TUI_clue = "bottom";
                    break;
                //------------------------------------------------------------
                case '8':
                    space.view_angle += space.rotation_step;
                    if(    space.loop_animation
                        || (space.frame_index < space.number_of_frames() - 1)
                      )
                        ++space;
                    space.TUI_clue = "+ show & orbit";
                    break;
                //------------------------------------------------------------
                case '*':
                    space.view_angle -= space.rotation_step;
                    if(    space.loop_animation
                        || (space.frame_index > 0)
                      )
                        --space;
                    space.TUI_clue = "- show & orbit";
                    break;
                //------------------------------------------------------------
                case '[':
                    space.current_frame().increment_egg();
                    space.TUI_clue = "+ start vertex  select";
                    break;
                //------------------------------------------------------------
                case '{':
                    space.current_frame().deccrement_egg();
                    space.TUI_clue = "- start vertex select";
                    break;
                //------------------------------------------------------------
                case ']':
                    space.current_frame().increment_spider();
                    space.TUI_clue = "+ end vertex  select";
                    break;
                //------------------------------------------------------------
                case '}':
                    space.current_frame().deccrement_spider();
                    space.TUI_clue = "- end vertex  select";
                    break;
                //------------------------------------------------------------
                case'\\':
                    space.current_frame().spider = (space.current_frame().number_of_vertices() - 1);
                    space.TUI_clue = "last vertex select";
                    break;
                //------------------------------------------------------------
                case '|':
                    space.current_frame().egg = 0;
                    space.TUI_clue = "first vertex select";
                    break;
                //------------------------------------------------------------
                case 'u':
                    space.current_frame().next_segment_select();
                    space.TUI_clue = "+ lit segment";
                    break;
                //------------------------------------------------------------
                case 'U':
                    space.current_frame().previous_segment_select();
                    space.TUI_clue = "- lit segment";
                    break;
                //------------------------------------------------------------
                case 'i':
                    space.current_frame().next_segment_egg();
                    space.TUI_clue = "+ start lit segment";
                    break;
                //------------------------------------------------------------
                case 'I':
                    space.current_frame().previous_segment_egg();
                    space.TUI_clue = "- start lit segment";
                    break;
                //------------------------------------------------------------
                case 'o':
                    space.current_frame().next_segment_spider();
                    space.TUI_clue = "+ end lit segment";
                    break;
                //------------------------------------------------------------
                case 'O':
                    space.current_frame().previous_segment_spider();
                    space.TUI_clue = "- end lit segment";
                    break;
                //------------------------------------------------------------
                case '/': // show frame set stats
                    space.show_frame_set_stats = !space.show_frame_set_stats;
                    break;
                //------------------------------------------------------------
                case '?': // show frame stats
                    space.show_frame_stats = !space.show_frame_stats;
                    break;
                //------------------------------------------------------------
                case '<':
                    space.current_frame().reverse_selected();
                    space.TUI_clue = "reversed selection";
                    break;
                //------------------------------------------------------------
                case 'c':
                    space.current_frame() += space.current_frame().selected_segment();
                    sprintf(formatted_string, "copied %d vertices to end of frame", (int)space.current_frame().selected_segment().size());
                    space.TUI_clue = formatted_string;
                    break;
                //------------------------------------------------------------
                case 'q':
                    offset = display_prompt_and_echo_int("set center X");
                    if(prompt_escaped)
                        break;
                    space.current_frame().move_selection(LaserBoy_3D_double(offset - space.current_frame().selected_segment().find_center().x,
                                                                            0,
                                                                            0
                                                                           )
                                                        );
                    break;
                //------------------------------------------------------------
                case 'w':
                    offset = display_prompt_and_echo_int("set center Y");
                    if(prompt_escaped)
                        break;
                    space.current_frame().move_selection(LaserBoy_3D_double(0,
                                                                            offset - space.current_frame().selected_segment().find_center().y,
                                                                            0
                                                                           )
                                                        );
                    break;
                //------------------------------------------------------------
                case 'e':
                    offset = display_prompt_and_echo_int("set center Z");
                    if(prompt_escaped)
                        break;
                    space.current_frame().move_selection(LaserBoy_3D_double(0,
                                                                            0,
                                                                            offset - space.current_frame().selected_segment().find_center().z
                                                                           )
                                                        );
                    break;
                //------------------------------------------------------------
                case 'x':
                    space.current_frame().move_selection(LaserBoy_3D_double(space.displacement_step, 0, 0));
                    space.TUI_clue = "+ X move";
                    break;
                //------------------------------------------------------------
                case 'X':
                    space.current_frame().move_selection(LaserBoy_3D_double(-space.displacement_step, 0, 0));
                    space.TUI_clue = "- X move";
                    break;
                //------------------------------------------------------------
                case 'y':
                    space.current_frame().move_selection(LaserBoy_3D_double(0, space.displacement_step, 0));
                    space.TUI_clue = "+ Y move";
                    break;
                //------------------------------------------------------------
                case 'Y':
                    space.current_frame().move_selection(LaserBoy_3D_double(0, -space.displacement_step, 0));
                    space.TUI_clue = "- Y move";
                    break;
                //------------------------------------------------------------
                case 'z':
                    space.current_frame().move_selection(LaserBoy_3D_double(0, 0, space.displacement_step));
                    space.TUI_clue = "+ Z move";
                    break;
                //------------------------------------------------------------
                case 'Z':
                    space.current_frame().move_selection(LaserBoy_3D_double(0, 0,-space.displacement_step));
                    space.TUI_clue = "- Z move";
                    break;
                //------------------------------------------------------------
                case 'f':
                    space.current_frame().rotate_selection(LaserBoy_3D_double(0, 0, space.rotation_step));
                    space.TUI_clue = "+ rotate Z";
                    break;
                //------------------------------------------------------------
                case 'F':
                    space.current_frame().rotate_selection(LaserBoy_3D_double(0, 0, -space.rotation_step));
                    space.TUI_clue = "- rotate Z";
                    break;
                //------------------------------------------------------------
                case 'g':
                    space.current_frame().rotate_selection(LaserBoy_3D_double(0, space.rotation_step, 0));
                    space.TUI_clue = "+ rotate y";
                    break;
                //------------------------------------------------------------
                case 'G':
                    space.current_frame().rotate_selection(LaserBoy_3D_double(0,-space.rotation_step, 0));
                    space.TUI_clue = "- rotate y";
                    break;
                //------------------------------------------------------------
                case 'r':
                    space.current_frame().rotate_selection_on_fulcrum(LaserBoy_3D_double(0, 0, space.rotation_step));
                    space.TUI_clue = "+ rotate Z on fulcrum";
                    break;
                //------------------------------------------------------------
                case 'R':
                    space.current_frame().rotate_selection_on_fulcrum(LaserBoy_3D_double(0, 0, -space.rotation_step));
                    space.TUI_clue = "- rotate Z on fulcrum";
                    break;
                //------------------------------------------------------------
                case 't':
                    space.current_frame().rotate_selection_on_fulcrum(LaserBoy_3D_double(0, space.rotation_step, 0));
                    space.TUI_clue = "+ rotate Y on fulcrum";
                    break;
                //------------------------------------------------------------
                case 'T':
                    space.current_frame().rotate_selection_on_fulcrum(LaserBoy_3D_double(0, -space.rotation_step, 0));
                    space.TUI_clue = "- rotate Y on fulcrum";
                    break;
                //------------------------------------------------------------
                case '9':
                    space.current_frame().scale_selection(LaserBoy_3D_double(space.magnitude_step_up, 1, 1));
                    space.TUI_clue = "+ scale X";
                    break;
                //------------------------------------------------------------
                case '(':
                    space.current_frame().scale_selection(LaserBoy_3D_double(space.magnitude_step_dn, 1, 1));
                    space.TUI_clue = "- scale X";
                    break;
                //------------------------------------------------------------
                case '0':
                    space.current_frame().scale_selection(LaserBoy_3D_double(1, space.magnitude_step_up, 1));
                    space.TUI_clue = "+ scale Y";
                    break;
                //------------------------------------------------------------
                case ')':
                    space.current_frame().scale_selection(LaserBoy_3D_double(1, space.magnitude_step_dn, 1));
                    space.TUI_clue = "- scale Y";
                    break;
                //------------------------------------------------------------
                case '-':
                    space.current_frame().scale_selection(LaserBoy_3D_double(1, 1, space.magnitude_step_up));
                    space.TUI_clue = "+ scale Z";
                    break;
                //------------------------------------------------------------
                case '_':
                    space.current_frame().scale_selection(LaserBoy_3D_double(1, 1, space.magnitude_step_dn));
                    space.TUI_clue = "- scale Z";
                    break;
                //------------------------------------------------------------
                case '=':
                    space.current_frame().scale_selection(LaserBoy_3D_double(space.magnitude_step_up, space.magnitude_step_up, space.magnitude_step_up));
                    space.TUI_clue = "+ scale XYZ";
                    break;
                //------------------------------------------------------------
                case '+':
                    space.current_frame().scale_selection(LaserBoy_3D_double(space.magnitude_step_dn, space.magnitude_step_dn, space.magnitude_step_dn));
                    space.TUI_clue = "- scale XYZ";
                    break;
                //------------------------------------------------------------
                case 'v':
                    space.current_frame().scale_selection_on_fulcrum(LaserBoy_3D_double(space.magnitude_step_up, 1, 1));
                    space.TUI_clue = "+ scale X by fulcrum";
                    break;
                //------------------------------------------------------------
                case 'V':
                    space.current_frame().scale_selection_on_fulcrum(LaserBoy_3D_double(space.magnitude_step_dn, 1, 1));
                    space.TUI_clue = "- scale X by fulcrum";
                    break;
                //------------------------------------------------------------
                case 'b':
                    space.current_frame().scale_selection_on_fulcrum(LaserBoy_3D_double(1, space.magnitude_step_up, 1));
                    space.TUI_clue = "+ scale Y by fulcrum";
                    break;
                //------------------------------------------------------------
                case 'B':
                    space.current_frame().scale_selection_on_fulcrum(LaserBoy_3D_double(1, space.magnitude_step_dn, 1));
                    space.TUI_clue = "- scale Y by fulcrum";
                    break;
                //------------------------------------------------------------
                case 'n':
                    space.current_frame().scale_selection_on_fulcrum(LaserBoy_3D_double(1, 1, space.magnitude_step_up));
                    space.TUI_clue = "+ size Z by fulcrum";
                    break;
                //------------------------------------------------------------
                case 'N':
                    space.current_frame().scale_selection_on_fulcrum(LaserBoy_3D_double(1, 1, space.magnitude_step_dn));
                    space.TUI_clue = "- scale Z by fulcrum";
                    break;
                //------------------------------------------------------------
                case 'm':
                    space.current_frame().scale_selection_on_fulcrum(LaserBoy_3D_double(space.magnitude_step_up, space.magnitude_step_up, space.magnitude_step_up));
                    space.TUI_clue = "+ scale XYZ by fulcrum";
                    break;
                //------------------------------------------------------------
                case 'M':
                    space.current_frame().scale_selection_on_fulcrum(LaserBoy_3D_double(space.magnitude_step_dn, space.magnitude_step_dn, space.magnitude_step_dn));
                    space.TUI_clue = "- scale XYZ by fulcrum";
                    break;
                //------------------------------------------------------------
                case 'a':
                    space.fulcrum.x += space.displacement_step;
                    space.TUI_clue = "+ X fulcrum";
                    break;
                //------------------------------------------------------------
                case 'A':
                    space.fulcrum.x -= space.displacement_step;
                    space.TUI_clue = "- X fulcrum";
                    break;
                //------------------------------------------------------------
                case 's':
                    space.fulcrum.y += space.displacement_step;
                    space.TUI_clue = "+ Y fulcrum";
                    break;
                //------------------------------------------------------------
                case 'S':
                    space.fulcrum.y -= space.displacement_step;
                    space.TUI_clue = "- Y fulcrum";
                    break;
                //------------------------------------------------------------
                case 'd':
                    space.fulcrum.z += space.displacement_step;
                    space.TUI_clue = "+ Z fulcrum";
                    break;
                //------------------------------------------------------------
                case 'D':
                    space.fulcrum.z -= space.displacement_step;
                    space.TUI_clue = "- Z fulcrum";
                    break;
                //------------------------------------------------------------
                case 'j':
                    temp_double = display_prompt_and_echo_double("set fulcrum X");
                    if(prompt_escaped)
                        break;
                    space.fulcrum.x = temp_double;
                    space.TUI_clue = "vertex X set";
                    break;
                //------------------------------------------------------------
                case 'k':
                    temp_double = display_prompt_and_echo_double("set fulcrum Y");
                    if(prompt_escaped)
                        break;
                    space.fulcrum.y = temp_double;
                    space.TUI_clue = "vertex Y set";
                    break;
                //------------------------------------------------------------
                case 'l':
                    temp_double = display_prompt_and_echo_double("set fulcrum Z");
                    if(prompt_escaped)
                        break;
                    space.fulcrum.z = temp_double;
                    space.TUI_clue = "vertex Z set";
                    break;
                //------------------------------------------------------------
                case ',':
                    space.fulcrum = space.current_frame().at(space.current_frame().egg);
                    space.TUI_clue = "fulcrum to egg";
                    break;
                //------------------------------------------------------------
                case '.':
                    space.fulcrum = space.current_frame().at(space.current_frame().spider);
                    space.TUI_clue = "fulcrum to spider";
                    break;
                //------------------------------------------------------------
                case ';': // toggle bitmap background
                    space.show_background_bitmap = !space.show_background_bitmap;
                    break;
                //------------------------------------------------------------
                  case SDLK_RETURN:
                    space.show_menu = !space.show_menu;
            } // end switch(event.key.keysym.unicode)
        } // end else if(event.type == SDL_KEYDOWN)
    } // end while(in_menu)
    space.show_menu = show_menu_was;
    return;
}
//############################################################################
void LaserBoy_SDL_GUI::wave_utilities_menu()
{
    bool   in_menu = true;
    string file_name;

    space.TUI_clue = "wave utilitites";

    while(in_menu)
    {
        space.render_space();
        space.overlay_wave_utilities_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    in_menu = false;
                    space.TUI_clue = "exit wave utilities";
                    break;
                //------------------------------------------------------------
                case '1': // wave utilities settings
                    wave_show_settings_menu();
                    break;
                //------------------------------------------------------------
                case '2': // show wave on screen
                    show_wave_options_menu();
                    break;
                //------------------------------------------------------------
                case '3':
                    space.overlay_file_list(LASERBOY_WAV_SHARE);
                    file_name = display_prompt_and_echo_name("enter LaserBoy wave file name to see stats");
                    if(prompt_escaped)
                        break;
                    space.TUI_clue = "wave stats";
                    space.render_space();

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_name))
                        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                    else
                        display_space();
                    wait_4_key();
                    break;
                //------------------------------------------------------------
                case '4':
                    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
                    file_name = display_prompt_and_echo_name("enter unformatted wave file name to see stats");
                    if(prompt_escaped)
                        break;
                    space.TUI_clue = "wave stats";
                    space.render_space();

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(!space.overlay_wave_stats(LASERBOY_UNFORMATTED_SHARE + file_name))
                        display_error(LASERBOY_UNFORMATTED_SHARE + file_name + " failed to open");
                    else
                        display_space();
                    wait_4_key();
                    break;
                //------------------------------------------------------------
                case '5': // set LaserBoy wave offsets
                    set_wave_offsets_menu();
                    break;
                //------------------------------------------------------------
                case '6': // apply LaserBoy wave sample offsets
                    space.overlay_file_list(LASERBOY_WAV_SHARE);
                    file_name = display_prompt_and_echo_name("enter wave file name to apply offsets");
                    if(prompt_escaped)
                        break;
                    space.TUI_clue = "offsetting wave samples";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(space.apply_wave_offsets_prep(LASERBOY_WAV_SHARE + file_name))
                    {
                        space.TUI_clue = "wave stats";
                        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_name))
                            display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                        else
                            display_space();
                    }
                    else
                        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");

                    wait_4_key();
                    break;
                //------------------------------------------------------------
                case '7': // overwrite LaserBoy wave sample offsets
                    space.overlay_file_list(LASERBOY_WAV_SHARE);
                    file_name = display_prompt_and_echo_name("enter wave file name to overwrite offsets");
                    if(prompt_escaped)
                        break;
                    space.TUI_clue = "overwriting offsets";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(space.overwirte_wave_offsets_prep(LASERBOY_WAV_SHARE + file_name))
                    {
                        space.TUI_clue = "wave stats";
                        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_name))
                            display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                        else
                            display_space();
                    }
                    else
                        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");

                    wait_4_key();
                    break;
                //------------------------------------------------------------
                case '8': // invert LaserBoy wave polarity
                    space.overlay_file_list(LASERBOY_WAV_SHARE);
                    file_name = display_prompt_and_echo_name("enter wave file name to globally invert");
                    if(prompt_escaped)
                        break;
                    space.TUI_clue = "inverting wave";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(space.invert_wave_prep(LASERBOY_WAV_SHARE + file_name))
                    {
                        space.TUI_clue = "wave stats";
                        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_name))
                            display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                        else
                            display_space();
                    }
                    else
                        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");

                    wait_4_key();
                    break;
                //------------------------------------------------------------
                case '9': // invert LaserBoy signal polarities
                    set_signal_polarities_menu();
                    break;
                //------------------------------------------------------------
                case '0': // clear LaserBoy wave polarity matrix
                    clear_wave_polarities_menu();
                    break;
                //------------------------------------------------------------
                case 'a': // impose byte value colors and on wave
                    black_level_to_wave_menu();
                    break;
                //------------------------------------------------------------
                case 'b': // impose bit resolution reduction on wave
                    bit_resolution_to_wave_menu();
                    break;
                //------------------------------------------------------------
                case 'c': // split wave
                    space.overlay_file_list(LASERBOY_WAV_SHARE);
                    file_name = display_prompt_and_echo_name("enter wave file name to split");
                    if(prompt_escaped)
                        break;
                    space.TUI_clue = "splitting wave";
                    space.render_space();
                    space.overlay_please_wait();
                    display_space();

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(space.split_wave_XY_rg_bi_LR_prep(   file_name
                                                          , LASERBOY_WAV_SHARE
                                                          , LASERBOY_UNFORMATTED_SHARE
                                                          , LASERBOY_AUDIO_SHARE
                                                        )
                      )
                    {
                        space.TUI_clue = "split wave";
                        space.render_space();
                        display_space();
                    }
                    else
                    {
                        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                        wait_4_key();
                    }
                    break;
                //------------------------------------------------------------
                case 'd': // join waves
                    join_XY_r_g_b_i_waves_menu();
                    break;
                //------------------------------------------------------------
                case 'e': // join waves
                    join_XY_rg_b_waves_menu();
                    break;
                //------------------------------------------------------------
                case 'f': // join waves
                    join_XY_rg_b_LR_waves_menu();
                    break;
                //------------------------------------------------------------
                case 'g': // add audio to wave
                    add_audio_to_wave_menu();
                    break;
            }
        }
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::wave_show_settings_menu()
{
    bool   in_menu = true;
    string file_name;

    space.TUI_clue = "wave show settings";

    while(in_menu)
    {
        space.render_space();
        space.overlay_wave_show_settings_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                case SDLK_ESCAPE:
                    space.TUI_clue = "action canceled";
                    in_menu = false;
                    break;
                //------------------------------------------------------------
                case '1':
                    space.wave_show_vertices = !space.wave_show_vertices;
                    if(space.wave_show_vertices)
                        space.TUI_clue = "show vertices";
                    else
                        space.TUI_clue = "hide vertices";
                    break;
                //------------------------------------------------------------
                case '2':
                    space.wave_show_blanks = !space.wave_show_blanks;
                    if(space.wave_show_blanks)
                        space.TUI_clue = "show blanking";
                    else
                        space.TUI_clue = "hide blanking";
                    break;
                //------------------------------------------------------------
                case '3':
                    space.wave_real_time = !space.wave_real_time;
                    if(space.wave_real_time)
                        space.TUI_clue = "real time";
                    else
                        space.TUI_clue = "as fast as possible";
                    break;
                //------------------------------------------------------------
                case '4':
                    space.wave_show_inverted = !space.wave_show_inverted;
                    if(space.wave_show_inverted)
                        space.TUI_clue = "inverted";
                    else
                        space.TUI_clue = "non-inverted";
                    break;
                //------------------------------------------------------------
                case '5':
                    space.wave_clean_screen = !space.wave_clean_screen;
                    if(space.wave_clean_screen)
                        space.TUI_clue = "clean screen";
                    else
                        space.TUI_clue = "no clean screen";
                    break;
                //------------------------------------------------------------
                case '6':
                    space.wave_loop = !space.wave_loop;
                    if(space.wave_loop)
                        space.TUI_clue = "loop";
                    else
                        space.TUI_clue = "no loop";
                    break;
            }
        }
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::show_wave_options_menu()
{
    string file_name;

    space.TUI_clue = "wave show settings";

    while(true)
    {
        space.render_space();
        space.overlay_show_wave_options_menu();
        display_space();

        SDL_WaitEvent(&event);
        if(event.type == SDL_KEYDOWN)
        {
            switch(event.key.keysym.unicode)
            {
                default:
                    space.TUI_clue = "bad key";
                    break;
                //------------------------------------------------------------
                case SDLK_ESCAPE:
                    space.TUI_clue = "action canceled";
                    return;
                //------------------------------------------------------------
                case '1': // show LaserBoy wave
                    space.overlay_file_list(LASERBOY_WAV_SHARE);
                    file_name = display_prompt_and_echo_name("enter wave file name to show");
                    if(prompt_escaped)
                        break;

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(display_wave(LASERBOY_WAV_SHARE + file_name))
                    {
                        while(space.wave_loop)
                        {
                            display_wave(LASERBOY_WAV_SHARE + file_name);
                            SDL_PollEvent(&event);
                            if(event.type == SDL_KEYDOWN)
                                break;
                        }
                    }
                    else
                        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                    break;
                //------------------------------------------------------------
                case '2': // show LaserBoy wave as unformatted
                    space.overlay_file_list(LASERBOY_WAV_SHARE);
                    file_name = display_prompt_and_echo_name("enter wave file name to show");
                    if(prompt_escaped)
                        break;

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(display_unframed_wave(LASERBOY_WAV_SHARE + file_name))
                    {
                        while(space.wave_loop)
                        {
                            display_unframed_wave(LASERBOY_WAV_SHARE + file_name);
                            SDL_PollEvent(&event);
                            if(event.type == SDL_KEYDOWN)
                                break;
                        }
                    }
                    else
                        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                    break;
                //------------------------------------------------------------
                case '3': // show unformatted wave
                    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
                    file_name = display_prompt_and_echo_name("enter wave file name to show");
                    if(prompt_escaped)
                        break;

                    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                        file_name += ".wav";

                    if(display_unframed_wave(LASERBOY_UNFORMATTED_SHARE + file_name))
                    {
                        while(space.wave_loop)
                        {
                            display_unframed_wave(LASERBOY_UNFORMATTED_SHARE + file_name);
                            SDL_PollEvent(&event);
                            if(event.type == SDL_KEYDOWN)
                                break;
                        }
                    }
                    else
                        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                    break;
                //------------------------------------------------------------
            }
            wait_4_key();
        }
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::set_wave_offsets_menu()
{
    string file_name;

    int temp_int;
    space.TUI_clue = "sample offsets";
    space.render_space();
    space.overlay_wave_offsets();
    display_space();

    temp_int = display_prompt_and_echo_int("enter wave sample offset for both X and Y");
    if(prompt_escaped)
        return;
    space.wave_offsets[0] = temp_int;
    space.wave_offsets[1] = temp_int;
    space.render_space();
    space.overlay_wave_offsets();
    display_space();
    if(prompt_escaped)
        return;
    space.render_space();
    space.overlay_wave_offsets();
    display_space();
    temp_int = display_prompt_and_echo_int("enter wave sample offset for red");
    if(prompt_escaped)
        return;
    space.wave_offsets[2] = temp_int;
    space.render_space();
    space.overlay_wave_offsets();
    display_space();
    temp_int = display_prompt_and_echo_int("enter wave sample offset for green");
    if(prompt_escaped)
        return;
    space.wave_offsets[3] = temp_int;
    space.render_space();
    space.overlay_wave_offsets();
    display_space();
    temp_int = display_prompt_and_echo_int("enter wave sample offset for blue");
    if(prompt_escaped)
        return;
    space.wave_offsets[4] = temp_int;
    space.render_space();
    space.overlay_wave_offsets();
    display_space();
    temp_int = display_prompt_and_echo_int("enter wave sample offset for auxiliary signal");
    if(prompt_escaped)
        return;
    space.wave_offsets[5] = temp_int;
    space.render_space();
    space.overlay_wave_offsets();
    display_space();
    temp_int = display_prompt_and_echo_int("enter wave sample offset for stereo audio");
    if(prompt_escaped)
        return;
    space.wave_offsets[6] = temp_int;
    space.wave_offsets[7] = temp_int;
    space.render_space();
    space.overlay_wave_offsets();
    display_space();

    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::set_signal_polarities_menu()
{
    short  wave_channels;
    string file_name;

    space.TUI_clue = "signal polarities";
    space.render_space();
    space.overlay_file_list(LASERBOY_WAV_SHARE);

    file_name = display_prompt_and_echo_name("enter wave file name to signal invert");
    if(prompt_escaped)
        return;

    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
        file_name += ".wav";

    if(!file_exists(LASERBOY_WAV_SHARE + file_name))
    {
        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
        wait_4_key();
        return;
    }

    for(int i = 0; i < 8; i++)
        space.signal_polarity[i] = -1; // not set!

    wave_channels = space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);

    if(wave_channels)
    {
        space.render_space();
        space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
        space.signal_polarity[0] = display_prompt_and_echo_bool("invert X ? (y | [any])");
        if(prompt_escaped)
            return;

        space.render_space();
        space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
        space.signal_polarity[1] = display_prompt_and_echo_bool("invert Y ? (y | [any])");
        if(prompt_escaped)
            return;

        space.render_space();
        space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
        space.signal_polarity[2] = display_prompt_and_echo_bool("invert red ? (y | [any])");
        if(prompt_escaped)
            return;

        space.render_space();
        space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
        space.signal_polarity[3] = display_prompt_and_echo_bool("invert green ? (y | [any])");
        if(prompt_escaped)
            return;

        space.render_space();
        space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
        space.signal_polarity[4] = display_prompt_and_echo_bool("invert blue ? (y | [any])");
        if(prompt_escaped)
            return;

        space.render_space();
        space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
        space.signal_polarity[5] = display_prompt_and_echo_bool("invert Z or other ? (y | [any])");
        if(prompt_escaped)
            return;

        if(wave_channels > 7)
        {
            space.render_space();
            space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
            space.signal_polarity[6] = display_prompt_and_echo_bool("invert L audio ? (y | [any])");
            if(prompt_escaped)
                return;

            space.render_space();
            space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
            space.signal_polarity[7] = display_prompt_and_echo_bool("invert R audio ? (y | [any])");
            if(prompt_escaped)
                return;
        }

        space.render_space();
        space.overlay_wave_signal_polarities(LASERBOY_WAV_SHARE + file_name, space.signal_polarity);
        display_space();
        SDL_Delay(250);

        space.TUI_clue = "inverting signals";
        space.overlay_please_wait();
        display_space();
        //--------------------------------------------------------------------
        space.invert_signals_prep(LASERBOY_WAV_SHARE + file_name);
        //--------------------------------------------------------------------
        space.TUI_clue = "wave stats";
        space.render_space();
        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_name))
            display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
        else
            display_space();
    }
    else
        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");

    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::clear_wave_polarities_menu()
{
    space.TUI_clue = "clear polarity matrix";
    space.render_space();
    space.overlay_file_list(LASERBOY_WAV_SHARE);
    string file_name = display_prompt_and_echo_name("enter wave file name to clear");
    if(prompt_escaped)
        return;
    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
        file_name += ".wav";
    if(!file_exists(LASERBOY_WAV_SHARE + file_name))
    {
        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    bool global_flip_flop = display_prompt_and_echo_bool("flip global polarity status? (y | [any])");
    if(prompt_escaped)
        return;
    space.TUI_clue = "clearing polarities";
    space.overlay_please_wait();
    display_space();
    if(space.clear_wave_polarity_list_prep(LASERBOY_WAV_SHARE + file_name, global_flip_flop))
    {
        space.TUI_clue = "wave stats";
        space.render_space();
        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_name))
            display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
        else
            display_space();
    }
    else
        display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::black_level_to_wave_menu()
{
    space.TUI_clue = "black level to wave";
    space.render_space();
    space.overlay_file_list(LASERBOY_WAV_SHARE);
    string file_in = display_prompt_and_echo_name("enter wave file name to black level");
    if(prompt_escaped)
        return;
    if(file_in.size() < 5 || file_in.substr(file_in.size() - 4) != ".wav")
        file_in += ".wav";
    if(!file_exists(LASERBOY_WAV_SHARE + file_in))
    {
        display_error(LASERBOY_WAV_SHARE + file_in + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_WAV_SHARE);
    string file_out = display_prompt_and_echo_name("enter name for output wave");
    if(prompt_escaped)
        return;
    if(file_out.size() < 5 || file_out.substr(file_out.size() - 4) != ".wav")
        file_out += ".wav";
    space.TUI_clue = "black leveling";
    space.overlay_please_wait();
    display_space();
    if(space.black_level_to_wave_prep(LASERBOY_WAV_SHARE + file_in, LASERBOY_WAV_SHARE + file_out))
    {
        space.TUI_clue = "wave stats";
        space.render_space();
        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_out))
            display_error(LASERBOY_WAV_SHARE + file_out + " failed to open");
        else
            display_space();
    }
    else
        display_error(LASERBOY_WAV_SHARE + file_out + " failed to open");
    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::bit_resolution_to_wave_menu()
{
    space.TUI_clue = "bit resolution to wave";
    space.render_space();
    space.overlay_file_list(LASERBOY_WAV_SHARE);
    string file_in = display_prompt_and_echo_name("enter wave file name to bit reduce");
    if(prompt_escaped)
        return;
    if(file_in.size() < 5 || file_in.substr(file_in.size() - 4) != ".wav")
        file_in += ".wav";
    if(!file_exists(LASERBOY_WAV_SHARE + file_in))
    {
        display_error(LASERBOY_WAV_SHARE + file_in + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_WAV_SHARE);
    string file_out = display_prompt_and_echo_name("enter name for output wave");
    if(prompt_escaped)
        return;
    if(file_out.size() < 5 || file_out.substr(file_out.size() - 4) != ".wav")
        file_out += ".wav";
    space.TUI_clue = "bit reducing";
    space.overlay_please_wait();
    display_space();
    if(space.bit_resolution_to_wave_prep(LASERBOY_WAV_SHARE + file_in, LASERBOY_WAV_SHARE + file_out))
    {
        space.TUI_clue = "wave stats";
        space.render_space();
        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_out))
            display_error(LASERBOY_WAV_SHARE + file_out + " failed to open");
        else
            display_space();
    }
    else
        display_error(LASERBOY_WAV_SHARE + file_out + " failed to open");
    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::join_XY_r_g_b_i_waves_menu()
{
    space.TUI_clue = "join XY r g b i";
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);

    string file_XY = display_prompt_and_echo_name("enter XY wave");
    if(prompt_escaped)
        return;
    if(file_XY.size()  < 5 || file_XY.substr(file_XY.size() - 4) != ".wav")
        file_XY  += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_XY))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_XY + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);

    string file_r = display_prompt_and_echo_name("enter red wave");
    if(prompt_escaped)
        return;
    if(file_r.size()   < 5 || file_r.substr(file_r.size() - 4) != ".wav")
        file_r   += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_r))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_r + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_g = display_prompt_and_echo_name("enter green wave");
    if(prompt_escaped)
        return;
    if(file_g.size()   < 5 || file_g.substr(file_g.size() - 4) != ".wav")
        file_g   += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_g))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_g + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_b = display_prompt_and_echo_name("enter blue wave");
    if(prompt_escaped)
        return;
    if(file_b.size()   < 5 || file_b.substr(file_b.size() - 4) != ".wav")
        file_b   += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_b))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_b + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_i = display_prompt_and_echo_name("enter intensity wave");
    if(prompt_escaped)
        return;
    if(file_i.size()   < 5 || file_i.substr(file_i.size() - 4) != ".wav")
        file_i   += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_i))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_i + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_out = display_prompt_and_echo_name("enter new name for wave out");
    if(prompt_escaped)
        return;

    bool global_polarity = display_prompt_and_echo_bool("is this wave inverted ? (y | [any])");
    if(prompt_escaped)
        return;

    space.TUI_clue = "joining waves";
    space.overlay_please_wait();
    display_space();
    if(file_out.size() < 5 || file_out.substr(file_out.size() - 4) != ".wav")
        file_out += ".wav";
    //------------------------------------------------------------------------
    if(space.join_XY_r_g_b_i_waves_prep(LASERBOY_UNFORMATTED_SHARE + file_XY,
                                        LASERBOY_UNFORMATTED_SHARE + file_r,
                                        LASERBOY_UNFORMATTED_SHARE + file_g,
                                        LASERBOY_UNFORMATTED_SHARE + file_b,
                                        LASERBOY_UNFORMATTED_SHARE + file_i,
                                        LASERBOY_WAV_SHARE         + file_out,
                                        global_polarity
                                       )
      )
    {
        space.TUI_clue = "joined waves";
        space.render_space();
        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_out))
            display_error(LASERBOY_WAV_SHARE + file_out + " failed to open");
        else
            display_space();
    }
    else
        display_error("at least one of those failed to open");
    //------------------------------------------------------------------------
    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::join_XY_rg_b_waves_menu()
{
    space.TUI_clue = "join XY rg b";
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_XY = display_prompt_and_echo_name("enter XY wave");
    if(prompt_escaped)
        return;
    if(file_XY.size()  < 5 || file_XY.substr(file_XY.size() - 4) != ".wav")
        file_XY  += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_XY))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_XY + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_rg = display_prompt_and_echo_name("enter red + green wave");
    if(prompt_escaped)
        return;
    if(file_rg.size()  < 5 || file_rg.substr(file_rg.size() - 4) != ".wav")
        file_rg   += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_rg))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_rg + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_b = display_prompt_and_echo_name("enter blue wave");
    if(prompt_escaped)
        return;
    if(file_b.size()   < 5 || file_b.substr(file_b.size() - 4) != ".wav")
        file_b   += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_b))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_b + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_out = display_prompt_and_echo_name("enter new name for wave out");
    if(prompt_escaped)
        return;
        
    bool global_polarity = display_prompt_and_echo_bool("is this wave inverted ? (y | [any])");
    if(prompt_escaped)
        return;

    space.TUI_clue = "joining waves";
    space.overlay_please_wait();
    display_space();
    if(file_out.size() < 5 || file_out.substr(file_out.size() - 4) != ".wav")
        file_out += ".wav";
    //------------------------------------------------------------------------
    if(space.join_XY_rg_b_waves_prep(LASERBOY_UNFORMATTED_SHARE + file_XY,
                                     LASERBOY_UNFORMATTED_SHARE + file_rg,
                                     LASERBOY_UNFORMATTED_SHARE + file_b,
                                     LASERBOY_WAV_SHARE         + file_out,
                                     global_polarity
                                    )
      )
    {
        space.TUI_clue = "joined waves";
        space.render_space();
        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_out))
            display_error(LASERBOY_WAV_SHARE + file_out + " failed to open");
        else
            display_space();
    }
    else
        display_error("something horrible and unexplainable just happened.");
    //------------------------------------------------------------------------
    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::join_XY_rg_b_LR_waves_menu()
{
    space.TUI_clue = "join XY rg b LR";
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_XY = display_prompt_and_echo_name("enter XY wave");
    if(prompt_escaped)
        return;
    if(file_XY.size()  < 5 || file_XY.substr(file_XY.size() - 4) != ".wav")
        file_XY  += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_XY))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_XY + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_rg = display_prompt_and_echo_name("enter red + green wave");
    if(prompt_escaped)
        return;
    if(file_rg.size()  < 5 || file_rg.substr(file_rg.size() - 4) != ".wav")
        file_rg   += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_rg))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_rg + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_b = display_prompt_and_echo_name("enter blue wave");
    if(prompt_escaped)
        return;
    if(file_b.size()   < 5 || file_b.substr(file_b.size() - 4) != ".wav")
        file_b   += ".wav";
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_b))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_b + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_AUDIO_SHARE);
    string file_LR = display_prompt_and_echo_name("enter stereo audio wave (_0 for silence)");
    if(prompt_escaped)
        return;
    if(    file_LR != "_0"
        && (    file_LR.size() < 5
             || file_LR.substr(file_LR.size() - 4) != ".wav"
           )
      )
        file_LR += ".wav";
    if(    file_LR != "_0"
        && !file_exists(LASERBOY_AUDIO_SHARE + file_LR)
      )
    {
        display_error(LASERBOY_AUDIO_SHARE + file_LR + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_out = display_prompt_and_echo_name("enter new name for wave out");
    if(prompt_escaped)
        return;

    bool global_polarity = display_prompt_and_echo_bool("is this wave inverted ? (y | [any])");
    if(prompt_escaped)
        return;

    space.TUI_clue = "joining waves";
    space.overlay_please_wait();
    display_space();
    if(file_out.size() < 5 || file_out.substr(file_out.size() - 4) != ".wav")
        file_out += ".wav";
    //------------------------------------------------------------------------
    if(space.join_XY_rg_b_LR_waves_prep(LASERBOY_UNFORMATTED_SHARE + file_XY,
                                        LASERBOY_UNFORMATTED_SHARE + file_rg,
                                        LASERBOY_UNFORMATTED_SHARE + file_b,
                                        LASERBOY_AUDIO_SHARE       + file_LR,
                                        LASERBOY_WAV_SHARE         + file_out,
                                        global_polarity
                                       )
      )
    {
        space.TUI_clue = "joined waves";
        space.render_space();
        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_out))
            display_error(LASERBOY_WAV_SHARE + file_out + " failed to open");
        else
            display_space();
    }
    else
        display_error("at least one of those failed to open");
    //------------------------------------------------------------------------
    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::add_audio_to_wave_menu()
{
    space.TUI_clue = "add audio to wave";
    space.render_space();
    space.overlay_file_list(LASERBOY_WAV_SHARE);
    string file_laser = display_prompt_and_echo_name("enter 6 channel laser wave");
    if(prompt_escaped)
        return;
    if(file_laser.size() < 5 || file_laser.substr(file_laser.size() - 4) != ".wav")
        file_laser += ".wav";
    if(!file_exists(LASERBOY_WAV_SHARE + file_laser))
    {
        display_error(LASERBOY_WAV_SHARE + file_laser + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_AUDIO_SHARE);
    string file_LR = display_prompt_and_echo_name("enter stereo audio wave (_0 for silence)");
    if(prompt_escaped)
        return;
    if(    file_LR != "_0"
        && (    file_LR.size() < 5
             || file_LR.substr(file_LR.size() - 4) != ".wav"
           )
      )
        file_LR += ".wav";
    if(    file_LR != "_0"
        && !file_exists(LASERBOY_AUDIO_SHARE + file_LR)
      )
    {
        display_error(LASERBOY_AUDIO_SHARE + file_LR + " failed to open");
        wait_4_key();
        return;
    }
    space.render_space();
    space.overlay_file_list(LASERBOY_WAV_SHARE);
    string file_out = display_prompt_and_echo_name("enter new name for wave out");
    if(prompt_escaped)
        return;
    space.TUI_clue = "adding audio";
    space.overlay_please_wait();
    display_space();
    //------------------------------------------------------------------------
    if(file_out.size() < 5 || file_out.substr(file_out.size() - 4) != ".wav")
        file_out += ".wav";
    //------------------------------------------------------------------------
    if(space.add_audio_to_wave_prep(LASERBOY_WAV_SHARE   + file_laser,
                                    LASERBOY_AUDIO_SHARE + file_LR,
                                    LASERBOY_WAV_SHARE   + file_out
                                   )
      )
    {
        space.TUI_clue = "added audio";
        space.render_space();
        if(!space.overlay_wave_stats(LASERBOY_WAV_SHARE + file_out))
            display_error(LASERBOY_WAV_SHARE + file_out + " failed to open");
        else
            display_space();
    }
    else
        display_error("at least one of those failed to open");
    //------------------------------------------------------------------------
    wait_4_key();
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::file_open_menu()
{
    bool bad_key = true;
    space.TUI_clue = "input file";
    while(bad_key)
    {
        space.render_space();
        space.overlay_in_file_type_menu();
        display_space();
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '1': // open ild
                if(ild_file_open_menu())
                    bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2': // open dxf
                if(dxf_file_open_menu())
                    bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3': // open wav
                if(wave_file_open_menu())
                    bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4': // open txt
                if(txt_file_open_menu())
                    bad_key = false;
                break;
            //----------------------------------------------------------------
            case '5': // open bmp
                if(bmp_file_open_menu())
                    bad_key = false;
                break;
        } // end switch(event.key.keysym.unicode)
    }
    return;
}

//############################################################################
bool LaserBoy_SDL_GUI::ild_file_open_menu()
{
    bool bad_key = true;
    string file_name;

    space.TUI_clue = "ild open";
    space.render_space();
    space.overlay_file_list(LASERBOY_ILD_SHARE);

    file_name = display_prompt_and_echo_name("enter ild file name");
    if(prompt_escaped)
        return false;

    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".ild")
        file_name += ".ild";

    while(bad_key)
    {
        space.overlay_in_ild_action_menu();
        display_space();
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                return false;
            //----------------------------------------------------------------
            case '1': // replace
                space.TUI_clue = "ild replace";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.from_ild_file(LASERBOY_ILD_SHARE + file_name)) // open failed
                {
                    display_error(LASERBOY_ILD_SHARE + file_name + " failed to open");
                    wait_4_key();
                    return false;
                }
                space.TUI_clue = LASERBOY_ILD_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2': // prepend
                space.TUI_clue = "ild prepend";
                space.render_space();
                space.overlay_please_wait();
                display_space();
                // scope
                {
                    LaserBoy_frame_set prepend(&space);
                    if(!prepend.from_ild_file(LASERBOY_ILD_SHARE + file_name))
                    {
                        display_error(LASERBOY_ILD_SHARE + file_name + " failed to open");
                        wait_4_key();
                        return false;
                    }
                    prepend += space;
                    space = prepend;
                }
                space.TUI_clue = LASERBOY_ILD_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3': // insert
                space.TUI_clue = "ild insert";
                space.render_space();
                space.overlay_please_wait();
                display_space();
                // scope
                {
                    int                i;
                    LaserBoy_frame_set insert(&space),
                                       result(&space);

                    if(!insert.from_ild_file(LASERBOY_ILD_SHARE + file_name))
                    {
                        display_error(LASERBOY_ILD_SHARE + file_name + " failed to open");
                        wait_4_key();
                        return false;
                    }
                    for(i = 0; i <= space.frame_index; i++)
                        result += space[i];
                    for(i = 0; i < insert.number_of_frames(); i++)
                        result += insert[i];
                    for(i = space.frame_index + 1; i < space.number_of_frames(); i++)
                        result += space[i];
                    space = result;
                }
                space.TUI_clue = LASERBOY_ILD_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4': // append
                space.TUI_clue = "ild append";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.from_ild_file(LASERBOY_ILD_SHARE + file_name, true))
                {
                    display_error(LASERBOY_ILD_SHARE + file_name + " failed to open");
                    wait_4_key();
                    return false;
                }
                space.TUI_clue = LASERBOY_ILD_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '5': // superimpose
                space.TUI_clue = "ild superimpose";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.superimpose_from_ild_file(LASERBOY_ILD_SHARE + file_name))
                {
                    display_error(LASERBOY_ILD_SHARE + file_name + " failed to open");
                    wait_4_key();
                    return false;
                }
                space.TUI_clue = LASERBOY_ILD_SHARE + file_name;
                bad_key = false;
                break;
        }
    }
    return true;
}

//############################################################################
bool LaserBoy_SDL_GUI::dxf_file_open_menu()
{
    bool   bad_key = true;
    string file_name;

    space.TUI_clue = "dxf open";

    while(bad_key)
    {
        space.render_space();
        space.overlay_in_dxf_action_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                return false;
            //----------------------------------------------------------------
            case '1': // open dxf
                space.overlay_file_list(LASERBOY_DXF_SHARE);
                file_name = display_prompt_and_echo_name("enter dxf file name");
                if(prompt_escaped)
                    break;

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".dxf")
                    file_name += ".dxf";

                space.TUI_clue = "dxf replace";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.from_dxf_file(LASERBOY_DXF_SHARE + file_name))
                {
                    display_error(LASERBOY_DXF_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_DXF_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2': // append dxf
                space.overlay_file_list(LASERBOY_DXF_SHARE);
                file_name = display_prompt_and_echo_name("enter dxf file name");
                if(prompt_escaped)
                    break;

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".dxf")
                    file_name += ".dxf";

                space.TUI_clue = "dxf append";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.from_dxf_file(LASERBOY_DXF_SHARE + file_name, true))
                {
                    display_error(LASERBOY_DXF_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_DXF_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3': // open dxf directory
                space.overlay_directory_list(LASERBOY_DXF_SHARE);
                file_name = display_prompt_and_echo_name("enter dxf directory name");
                if(prompt_escaped)
                    break;
                space.TUI_clue = "dxf set replace";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.from_dxf_directory(LASERBOY_DXF_SHARE + file_name))
                {
                    display_error(LASERBOY_DXF_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_DXF_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4': // append dxf directory
                space.overlay_directory_list(LASERBOY_DXF_SHARE);
                file_name = display_prompt_and_echo_name("enter dxf directory name");
                if(prompt_escaped)
                    break;
                space.TUI_clue = "dxf set append";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.from_dxf_directory(LASERBOY_DXF_SHARE + file_name, true))
                {
                    display_error(LASERBOY_DXF_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_DXF_SHARE + file_name;
                bad_key = false;
                break;
        }
    }
    return true;
}

//############################################################################
bool LaserBoy_SDL_GUI::wave_file_open_menu()
{
    bool   bad_key = true;
    string file_name;

    space.TUI_clue = "open wave";

    while(bad_key)
    {
        space.render_space();
        space.overlay_in_wave_action_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                return false;
            //----------------------------------------------------------------
            case '1': // replace frame set with LaserBoy formatted wave
                space.overlay_file_list(LASERBOY_WAV_SHARE);
                file_name = display_prompt_and_echo_name("enter wave file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "opening wave";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                    file_name += ".wav";

                if(!space.from_LaserBoy_wave_file(LASERBOY_WAV_SHARE + file_name))
                {
                    display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_WAV_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2': // append frame set with LaserBoy formatted wave
                space.overlay_file_list(LASERBOY_WAV_SHARE);
                file_name = display_prompt_and_echo_name("enter wave file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "opening wave";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                    file_name += ".wav";

                if(!space.from_LaserBoy_wave_file(LASERBOY_WAV_SHARE + file_name, true))
                {
                    display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_WAV_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3': // open LaserBoy wave as unformatted
                space.overlay_file_list(LASERBOY_WAV_SHARE);
                file_name = display_prompt_and_echo_name("enter wave file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "opening wave";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                    file_name += ".wav";

                if(!space.from_unformatted_wave_file(LASERBOY_WAV_SHARE + file_name))
                {
                    display_error(LASERBOY_WAV_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_WAV_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4': // replace frame set with unformatted wave
                space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
                file_name = display_prompt_and_echo_name("enter wave file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "opening wave";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                    file_name += ".wav";

                if(!space.from_unformatted_wave_file(LASERBOY_UNFORMATTED_SHARE + file_name))
                {
                    display_error(LASERBOY_UNFORMATTED_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_UNFORMATTED_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '5': // append frame set with unformatted wave
                space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
                file_name = display_prompt_and_echo_name("enter wave file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "opening wave";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                    file_name += ".wav";

                if(!space.from_unformatted_wave_file(LASERBOY_UNFORMATTED_SHARE + file_name, true))
                {
                    display_error(LASERBOY_UNFORMATTED_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = LASERBOY_UNFORMATTED_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '6': // open qm blanked wave
                if(wav_qm_open_menu(false))
                    bad_key = false;
                break;
            //----------------------------------------------------------------
            case '7': // open qm blanked wave
                if(wav_qm_open_menu(true))
                    bad_key = false;
                break;
        }
    }
    return true;
}


//############################################################################
bool LaserBoy_SDL_GUI::wav_qm_open_menu(bool append)
{
    space.overlay_file_list(LASERBOY_UNFORMATTED_SHARE);
    string file_name = display_prompt_and_echo_name("enter wave file name");
    if(prompt_escaped)
        return false;

    space.TUI_clue = "opening wave";
    space.render_space();
    space.overlay_please_wait();
    display_space();

    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
        file_name += ".wav";
        
    if(!file_exists(LASERBOY_UNFORMATTED_SHARE + file_name))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_name + " failed to open");
        wait_4_key();
        return false;
    }

    bool global_polarity = display_prompt_and_echo_bool("is this wave inverted ? (y | [any])");
    if(prompt_escaped)
        return false;

    if(!space.from_qm_wave_file(LASERBOY_UNFORMATTED_SHARE + file_name, global_polarity, append))
    {
        display_error(LASERBOY_UNFORMATTED_SHARE + file_name + " failed to open");
        wait_4_key();
        return false;
    }
    space.TUI_clue = LASERBOY_UNFORMATTED_SHARE + file_name;
    return true;
}

//############################################################################
bool LaserBoy_SDL_GUI::txt_file_open_menu()
{
    bool bad_key = true;
    string file_name;

    space.TUI_clue = "txt open";
    space.render_space();
    space.overlay_file_list(LASERBOY_TXT_SHARE);

    file_name = display_prompt_and_echo_name("enter ASCII txt file name");
    if(prompt_escaped)
        return false;

    if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".txt")
        file_name += ".txt";

    while(bad_key)
    {
        space.overlay_in_txt_action_menu();
        display_space();
        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                return false;
            //----------------------------------------------------------------
            case '1': // replace
                space.TUI_clue = "txt replace";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.from_txt_file(LASERBOY_TXT_SHARE + file_name)) // open failed
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to open");
                    wait_4_key();
                    return false;
                }
                space.TUI_clue = LASERBOY_TXT_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2': // prepend
                space.TUI_clue = "txt prepend";
                space.render_space();
                space.overlay_please_wait();
                display_space();
                // scope
                {
                    LaserBoy_frame_set prepend(&space);
                    if(!prepend.from_txt_file(LASERBOY_TXT_SHARE + file_name))
                    {
                        display_error(LASERBOY_TXT_SHARE + file_name + " failed to open");
                        wait_4_key();
                        return false;
                    }
                    prepend += space;
                    space = prepend;
                }
                space.TUI_clue = LASERBOY_TXT_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3': // insert
                space.TUI_clue = "txt insert";
                space.render_space();
                space.overlay_please_wait();
                display_space();
                // scope
                {
                    int                i;
                    LaserBoy_frame_set insert(&space),
                                       result(&space);

                    if(!insert.from_txt_file(LASERBOY_TXT_SHARE + file_name))
                    {
                        display_error(LASERBOY_TXT_SHARE + file_name + " failed to open");
                        wait_4_key();
                        return false;
                    }
                    for(i = 0; i <= space.frame_index; i++)
                        result += space[i];
                    for(i = 0; i < insert.number_of_frames(); i++)
                        result += insert[i];
                    for(i = space.frame_index + 1; i < space.number_of_frames(); i++)
                        result += space[i];
                    space = result;
                }
                space.TUI_clue = LASERBOY_TXT_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4': // append
                space.TUI_clue = "txt append";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.from_txt_file(LASERBOY_TXT_SHARE + file_name, true))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to open");
                    wait_4_key();
                    return false;
                }
                space.TUI_clue = LASERBOY_TXT_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '5': // superimpose
                space.TUI_clue = "txt superimpose";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.superimpose_from_txt_file(LASERBOY_TXT_SHARE + file_name))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to open");
                    wait_4_key();
                    return false;
                }
                space.TUI_clue = LASERBOY_TXT_SHARE + file_name;
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '6': // superimpose
                space.TUI_clue = "txt palettes only";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.palettes_from_txt_file(LASERBOY_TXT_SHARE + file_name))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to open");
                    wait_4_key();
                    return false;
                }
                space.TUI_clue = LASERBOY_TXT_SHARE + file_name;
                bad_key = false;
                break;
        }
    }
    return true;
}

//############################################################################
bool LaserBoy_SDL_GUI::bmp_file_open_menu()
{
    bool bad_key = true;
    string file_name;

    space.TUI_clue = "bmp open";

    while(bad_key)
    {
        space.render_space();
        space.overlay_in_bmp_action_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                return false;
            //----------------------------------------------------------------
            case '1':
                space.overlay_file_list(LASERBOY_BMP_SHARE);
                file_name = display_prompt_and_echo_name("enter bitmap file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "loading bmp";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".bmp")
                    file_name += ".bmp";

                if(!space.current_frame().color_from_bmp(LASERBOY_BMP_SHARE + file_name))
                {
                    display_error(LASERBOY_BMP_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = "color over bmp frame";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2':
                space.overlay_file_list(LASERBOY_BMP_SHARE);
                file_name = display_prompt_and_echo_name("enter bitmap file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "loading bmp";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".bmp")
                    file_name += ".bmp";

                if(!space.color_from_bmp(LASERBOY_BMP_SHARE + file_name))
                {
                    display_error(LASERBOY_BMP_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = "color over bmp all";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3':
                space.overlay_file_list(LASERBOY_BMP_SHARE);
                file_name = display_prompt_and_echo_name("enter bitmap file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "loading bmp";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".bmp")
                    file_name += ".bmp";

                if(!space.current_frame().subtract_bmp(LASERBOY_BMP_SHARE + file_name))
                {
                    display_error(LASERBOY_BMP_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = "subtract bmp frame";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4':
                space.overlay_file_list(LASERBOY_BMP_SHARE);
                file_name = display_prompt_and_echo_name("enter bitmap file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "loading bmp";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".bmp")
                    file_name += ".bmp";

                if(!space.subtract_bmp(LASERBOY_BMP_SHARE + file_name))
                {
                    display_error(LASERBOY_BMP_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = "subtract bmp all";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '5':
                space.overlay_file_list(LASERBOY_BMP_SHARE);
                file_name = display_prompt_and_echo_name("enter bitmap file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "loading bmp";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".bmp")
                    file_name += ".bmp";

                if(!space.load_background_bitmap(LASERBOY_BMP_SHARE + file_name))
                {
                    display_error(LASERBOY_BMP_SHARE + file_name + " failed to open");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = "background bmp";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '6':
                space.clear_background_bitmap();
                space.TUI_clue = "clear background";
                bad_key = false;
                break;
        }
    }
    return true;
}

//############################################################################
void LaserBoy_SDL_GUI::save_as_file_menu()
{
    bool   bad_key = true;
    string file_name;

    space.TUI_clue = "save ild";

    while(bad_key)
    {
        space.render_space();
        space.overlay_out_file_type_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '1':
                save_as_ild_menu();
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2':
                save_as_dxf_menu();
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3':
                save_as_wav_menu();
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4':
                save_as_txt_menu();
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '5':
                save_as_bmp_menu();
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '6':
                space.save_cfg_file();
                space.TUI_clue = "LaserBoy.wtf saved";
                bad_key = false;
                break;
        } // end switch(event.key.keysym.unicode)
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::save_as_ild_menu()
{
    bool   bad_key = true;
    string file_name;

    space.TUI_clue = "save as file";

    while(bad_key)
    {
        space.render_space();
        space.overlay_out_ild_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '1':
                file_name = display_prompt_and_echo_name("enter new ild file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving ild";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".ild")
                    file_name += ".ild";

                if(!space.current_frame().save_as_ild(LASERBOY_ILD_SHARE + file_name))
                {
                    display_error(LASERBOY_ILD_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "ild saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2':
                file_name = display_prompt_and_echo_name("enter new ild file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving ild";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".ild")
                    file_name += ".ild";

                if(!space.selected_frames().save_as_ild(LASERBOY_ILD_SHARE + file_name))
                {
                    display_error(LASERBOY_ILD_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "ild saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3':
                file_name = display_prompt_and_echo_name("enter new ild file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving ild";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".ild")
                    file_name += ".ild";

                if(!space.save_as_ild(LASERBOY_ILD_SHARE + file_name))
                {
                    display_error(LASERBOY_ILD_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "ild saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4':
                space.overlay_frame_effects();
                file_name = display_prompt_and_echo_name("enter frame effect name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving ild";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.call_frame_effect(file_name))
                {
                    display_error(file_name + " is not a registered LaserBoy frame effect");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "ild saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '5':
                space.overlay_frame_set_effects();
                file_name = display_prompt_and_echo_name("enter frame set effect name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving ild";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.call_frame_set_effect(file_name))
                {
                    display_error(file_name + " is not a registered LaserBoy frame set effect");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "ild saved";
                bad_key = false;
                break;
        } // end switch(event.key.keysym.unicode)
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::save_as_dxf_menu()
{
    bool   bad_key = true;
    string file_name;

    space.TUI_clue = "save dxf";

    while(bad_key)
    {
        space.render_space();
        space.overlay_out_dxf_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '1':
                file_name = display_prompt_and_echo_name("enter new dxf file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving dxf";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".dxf")
                    file_name += ".dxf";

                if(!space.current_frame().save_as_dxf(LASERBOY_DXF_SHARE + file_name))
                {
                    display_error(LASERBOY_DXF_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "dxf saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2':
                file_name = display_prompt_and_echo_name("enter new dxf directory name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving dxf";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.selected_frames().save_as_dxf_directory(LASERBOY_DXF_SHARE + file_name))
                {
                    display_error(LASERBOY_DXF_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "dxf directory saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3':
                file_name = display_prompt_and_echo_name("enter new dxf directory name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving dxf";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.save_as_dxf_directory(LASERBOY_DXF_SHARE + file_name))
                {
                    display_error(LASERBOY_DXF_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "dxf directory saved";
                bad_key = false;
                break;
        } // end switch(event.key.keysym.unicode)
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::save_as_wav_menu()
{
    bool    bad_key = true;
    string  file_name;

    space.TUI_clue = "save as file";

    while(bad_key)
    {
        space.render_space();
        space.overlay_out_wav_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '1':
                file_name = display_prompt_and_echo_name("enter new wave file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving wave";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                    file_name += ".wav";

                if(!space.save_as_wave(LASERBOY_WAV_SHARE + file_name))
                {
                    display_error(LASERBOY_WAV_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "wave saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2':
                file_name = display_prompt_and_echo_name("enter new wave file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving wave";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".wav")
                    file_name += ".wav";

                if(!space.current_frame().save_as_wave(LASERBOY_WAV_SHARE + file_name))
                {
                    display_error(LASERBOY_WAV_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "wave saved";
                bad_key = false;
                break;
        } // end switch(event.key.keysym.unicode)
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::save_as_txt_menu()
{
    bool   bad_key = true;
    string file_name;

    space.TUI_clue = "save ASCII";

    while(bad_key)
    {
        space.render_space();
        space.overlay_out_txt_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '1':
                file_name = display_prompt_and_echo_name("enter new ild file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving txt";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".txt")
                    file_name += ".txt";

                if(!space.current_frame().save_as_txt(LASERBOY_TXT_SHARE + file_name))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "txt saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2':
                file_name = display_prompt_and_echo_name("enter new txt file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving txt";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".txt")
                    file_name += ".txt";

                if(!space.selected_frames().save_as_txt(LASERBOY_TXT_SHARE + file_name))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "txt saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3':
                file_name = display_prompt_and_echo_name("enter new txt file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving txt";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".txt")
                    file_name += ".txt";

                if(!space.save_as_txt(LASERBOY_TXT_SHARE + file_name))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "txt saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '4':
                file_name = display_prompt_and_echo_name("enter new palette file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving txt";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".txt")
                    file_name += ".txt";

                if(space.current_frame().palette_index == LASERBOY_TRUE_COLOR)
                {
                    if(!space.current_frame().save_as_txt_table(LASERBOY_TXT_SHARE + file_name))
                    {
                        display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                        wait_4_key();
                        break;
                    }
                }
                else if(!space.palette_picker(space.current_frame().palette_index).save_as_txt(LASERBOY_TXT_SHARE + file_name))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = "txt saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '5':
                file_name = display_prompt_and_echo_name("enter new palette file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving txt";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".txt")
                    file_name += ".txt";

                if(!space.palette_picker(space.target_palette_index).save_as_txt(LASERBOY_TXT_SHARE + file_name))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }
                space.TUI_clue = "txt saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '6':
                file_name = display_prompt_and_echo_name("enter new C++ file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving C++";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.current_frame().save_as_code(file_name, LASERBOY_TXT_SHARE))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "C++ saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '7':
                file_name = display_prompt_and_echo_name("enter new C++ file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving C++";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.selected_frames().save_as_code(file_name, LASERBOY_TXT_SHARE))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "C++ saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '8':
                file_name = display_prompt_and_echo_name("enter new C++ file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving C++";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(!space.save_as_code(file_name, LASERBOY_TXT_SHARE))
                {
                    display_error(LASERBOY_TXT_SHARE + file_name + " failed to save");
                    wait_4_key();
                    break;
                }

                space.TUI_clue = "C++ saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '9':
                space.save_cfg_file();
                space.TUI_clue = "LaserBoy.wtf saved";
                bad_key = false;
                break;
        } // end switch(event.key.keysym.unicode)
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::save_as_bmp_menu()
{
    bool   bad_key = true;
    string file_name;

    space.TUI_clue = "save bmp";

    while(bad_key)
    {
        space.render_space();
        space.overlay_out_bmp_menu();
        display_space();

        do
        {
            SDL_WaitEvent(&event);
        } while(event.type != SDL_KEYDOWN);

        switch(event.key.keysym.unicode)
        {
            default:
                space.TUI_clue = "bad key";
                break;
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                space.TUI_clue = "action canceled";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '1':
                file_name = display_prompt_and_echo_name("enter new bitmap file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving bitmap";
                space.render_space();
                space.overlay_please_wait();
                display_space();

                if(file_name.size() < 5 || file_name.substr(file_name.size() - 4) != ".bmp")
                    file_name += ".bmp";

                space.current_frame().save_as_bmp(LASERBOY_BMP_SHARE + file_name);
                space.TUI_clue = "bitmap saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '2':
                file_name = display_prompt_and_echo_name("enter new bitmap directory file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving bitmap directory";
                space.render_space();
                space.overlay_please_wait();
                display_space();
                space.selected_frames().save_as_bmp_directory(LASERBOY_BMP_SHARE + file_name);
                space.TUI_clue = "bitmap directory saved";
                bad_key = false;
                break;
            //----------------------------------------------------------------
            case '3':
                file_name = display_prompt_and_echo_name("enter new bitmap directory file name");
                if(prompt_escaped)
                    break;

                space.TUI_clue = "saving bitmap directory";
                space.render_space();
                space.overlay_please_wait();
                display_space();
                space.save_as_bmp_directory(LASERBOY_BMP_SHARE + file_name);
                space.TUI_clue = "bitmap directory saved";
                bad_key = false;
                break;
        } // end switch(event.key.keysym.unicode)
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::process_terminate_request()
{
    LaserBoy_color bg_was = space.bg_color;
    space.bg_color = (u_int)0x00ff0000;
    space.recolor_background();
    space.TUI_clue = "terminating";
    space.render_space();
    display_space();

    if(display_prompt_and_echo_bool("Do you really want to terminate LaserBoy? (y | [any])"))
        running = false;
    else
    {
        space.bg_color = bg_was;
        space.recolor_background();
        teminate_requested = false;
        space.TUI_clue = "terminate canceled";
        space.render_space();
        if(space.show_menu)
            space.overlay_program_menu();
        display_space();
    }
    return;
}

//############################################################################
void LaserBoy_SDL_GUI::Laserboy_terminating()
{
    space.bg_color = (u_int)0x000000ff;
    space.recolor_background();
    space.TUI_clue = "terminating";
    space.render_space();
    display_space();
    //------------------------------------------------------------------------
    if(!display_prompt_and_echo_bool("Exit without saving OUT.ild? (y | [any])"))
    {
        space.TUI_clue = "saving OUT.ild";
        space.render_space();
        display_space();
        if(space.auto_optimize)
            space.minimize();
        space.save_as_ild(LASERBOY_ILD_SHARE + "OUT.ild");
    }
    //------------------------------------------------------------------------
    else
    {
        space.TUI_clue = "LaserBoy terminated";
        space.render_space();
        display_space();
    }
    //------------------------------------------------------------------------
    SDL_Delay(500);
}

//############################################################################
void LaserBoy_SDL_GUI::start_menu_loop()
{
    char formatted_string[80] = {0};
    running = true;

    space.TUI_clue = "NO DATA!";
    if(space.from_ild_file(LASERBOY_ILD_SHARE + "in.ild"))
        space.TUI_clue = LASERBOY_ILD_SHARE + "in.ild";
    display_space();

    while(running)
    {
        space.render_space();
        if(space.show_menu)
            space.overlay_program_menu();
        if(space.show_frame_stats)
            space.overlay_frame_stats();
        if(space.show_frame_set_stats)
            space.overlay_frame_set_stats();
        display_space();
        //--------------------------------------------------------------------
        do
        {
            SDL_WaitEvent(&event);
        } while((event.type != SDL_KEYDOWN));

        switch(event.key.keysym.unicode)
        {
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                teminate_requested = true;
                break;
            //----------------------------------------------------------------
            case SDLK_RETURN:
                space.show_menu = !space.show_menu;
                if(!space.show_menu)
                    space.TUI_clue = "[Enter] help";
                break;
            //----------------------------------------------------------------
            default:
                check_for_SDL_arrow_keys(event.key.keysym.sym);
                break;
            //----------------------------------------------------------------
            case SDLK_BACKSPACE:
                space.delete_current_frame();
                space.TUI_clue = "deleted";
                break;
            //----------------------------------------------------------------
            case '`': // start animation forward
                animate_forward();
                break;
            //----------------------------------------------------------------
            case '~': // start animation reverse
                animate_reverse();
                break;
            //----------------------------------------------------------------
            case '-': // start rotation formard
                rotate_forward();
                break;
            //----------------------------------------------------------------
            case '_': // start rotation reverse
                rotate_reverse();
                break;
            //----------------------------------------------------------------
            case '=': // start rotation & animation formard
                animate_rotate_forward();
                break;
            //----------------------------------------------------------------
            case '+': // start rotation and animation reverse
                animate_rotate_reverse();
                break;
            //----------------------------------------------------------------
            case '/': // show frame set stats
                space.show_frame_set_stats = !space.show_frame_set_stats;
                break;
            //----------------------------------------------------------------
            case '?': // show frame stats
                space.show_frame_stats = !space.show_frame_stats;
                break;
            //----------------------------------------------------------------
            case ';': // toggle bitmap background
                space.show_background_bitmap = !space.show_background_bitmap;
                break;
            //----------------------------------------------------------------
            case '1':
                space.view_angle.x += space.rotation_step;
                space.TUI_clue = "+ orbit X";
                break;
            //----------------------------------------------------------------
            case '!':
                space.view_angle.x -= space.rotation_step;
                space.TUI_clue = "- orbit X";
                break;
            //----------------------------------------------------------------
            case '2':
                space.view_angle.y += space.rotation_step;
                space.TUI_clue = "+ orbit Y";
                break;
            //----------------------------------------------------------------
            case '@':
                space.view_angle.y -= space.rotation_step;
                space.TUI_clue = "- orbit Y";
                break;
            //----------------------------------------------------------------
            case '3':
                space.view_angle.z += space.rotation_step;
                space.TUI_clue = "+ orbit Z";
                break;
            //----------------------------------------------------------------
            case '#':
                space.view_angle.z -= space.rotation_step;
                space.TUI_clue = "- orbit Z";
                break;
            //----------------------------------------------------------------
            case '4':
                space.view_angle += space.rotation_step;
                space.TUI_clue = "+ orbit XYZ";
                break;
            //----------------------------------------------------------------
            case '$':
                space.view_angle -= space.rotation_step;
                space.TUI_clue = "- orbit XYZ";
                break;
            //----------------------------------------------------------------
            case '5':
                space.view_angle   = 0;
                space.TUI_clue = "front";
                break;
            //----------------------------------------------------------------
            case '%':
                space.view_angle.x = 0;
                space.view_angle.y = pi;
                space.view_angle.z = 0;
                space.TUI_clue = "back";
                break;
            //----------------------------------------------------------------
            case '6':
                space.view_angle.x = 0;
                space.view_angle.y = half_pi;
                space.view_angle.z = 0;
                space.TUI_clue = "left side";
                break;
            //----------------------------------------------------------------
            case '^':
                space.view_angle.x = 0;
                space.view_angle.y = -half_pi;
                space.view_angle.z = 0;
                space.TUI_clue = "right side";
                break;
            //----------------------------------------------------------------
            case '7':
                space.view_angle.x = half_pi;
                space.view_angle.y = 0;
                space.view_angle.z = 0;
                space.TUI_clue = "top";
                break;
            //----------------------------------------------------------------
            case '&':
                space.view_angle.x = -half_pi;
                space.view_angle.y = 0;
                space.view_angle.z = 0;
                space.TUI_clue = "bottom";
                break;
            //----------------------------------------------------------------
            case '8':
                space.view_angle += space.rotation_step;
                if(    space.loop_animation
                    || (space.frame_index < space.number_of_frames() - 1)
                  )
                    ++space;
                space.TUI_clue = "+ frame & orbit";
                break;
            //----------------------------------------------------------------
            case '*':
                space.view_angle -= space.rotation_step;
                if(    space.loop_animation
                    || (space.frame_index > 0)
                  )
                    --space;
                space.TUI_clue = "- frame & orbit";
                break;
            //----------------------------------------------------------------
            case '9': // go to first frame
                space.frame(0);
                space.TUI_clue = "first frame";
                break;
            //----------------------------------------------------------------
            case '0': // go to last frame
                space.frame(space.number_of_frames() - 1);
                space.TUI_clue = "last frame";
                break;
            //----------------------------------------------------------------
            case '(': // set, unset start frame select
                space.toggle_frame_select_start();
                if(space.frame_select_start != -1)
                {
                    sprintf(formatted_string, "start = %d", space.frame_select_start);
                    space.TUI_clue = formatted_string;
                }
                else
                    space.TUI_clue = "start unset";
                break;
            //----------------------------------------------------------------
            case ')': // set, unset end frame select
                space.toggle_frame_select_end();
                if(space.frame_select_end != -1)
                {
                    sprintf(formatted_string, "end = %d", space.frame_select_end);
                    space.TUI_clue = formatted_string;
                }
                else
                    space.TUI_clue = "end unset";
                break;
            //----------------------------------------------------------------
            case ' ': // [Space] select, deselect current frame
                space.current_frame().is_selected = !(space.current_frame().is_selected);
                if(space.current_frame().is_selected)
                    space.TUI_clue = "select frame";
                else
                    space.TUI_clue = "deselect frame";
                break;
            //----------------------------------------------------------------
            case 'q':
                space.view_offset.x += space.displacement_step;
                space.TUI_clue = "+ pan X";
                break;
            //----------------------------------------------------------------
            case 'Q':
                space.view_offset.x -= space.displacement_step;
                space.TUI_clue = "- pan X";
                break;
            //----------------------------------------------------------------
            case 'w':
                space.view_offset.y += space.displacement_step;
                space.TUI_clue = "+ pan Y";
                break;
            //----------------------------------------------------------------
            case 'W':
                space.view_offset.y -= space.displacement_step;
                space.TUI_clue = "- pan Y";
                break;
            //----------------------------------------------------------------
            case 'e':
                space.view_offset += space.displacement_step;
                space.TUI_clue = "+ pan XY";
                break;
            //----------------------------------------------------------------
            case 'E':
                space.view_offset -= space.displacement_step;
                space.TUI_clue = "- pan XY";
                break;
            //----------------------------------------------------------------
            case 'r':
            case 'R':
                space.view_offset = 0.0;
                space.TUI_clue = "on center";
                break;
            //----------------------------------------------------------------
            case 'a':
                space.view_scale.x *= space.magnitude_step_up;
                space.TUI_clue = "+ zoom X";
                break;
            //----------------------------------------------------------------
            case 'A':
                space.view_scale.x *= space.magnitude_step_dn;
                space.TUI_clue = "- zoom X";
                break;
            //----------------------------------------------------------------
            case 's':
                space.view_scale.y *= space.magnitude_step_up;
                space.TUI_clue = "+ zoom Y";
                break;
            //----------------------------------------------------------------
            case 'S':
                space.view_scale.y *= space.magnitude_step_dn;
                space.TUI_clue = "- zoom Y";
                break;
            //----------------------------------------------------------------
            case 'd':
                space.view_scale *= space.magnitude_step_up;
                space.TUI_clue = "+ zoom XY";
                break;
            //----------------------------------------------------------------
            case 'D':
                space.view_scale *= space.magnitude_step_dn;
                space.TUI_clue = "- zoom XY";
                break;
            //----------------------------------------------------------------
            case 'f':
            case 'F':
                space.view_scale = 1.00;
                space.TUI_clue = "no zoom";
                break;
            //----------------------------------------------------------------
            case SDLK_TAB:
                system_settings_menu();
                break;
            //----------------------------------------------------------------
            case 't': // signal bit resolution
            case 'T':
                signal_bit_resolution();
                break;
            //----------------------------------------------------------------
            case 'u': // UI visuals
            case 'U':
                visuals_settings_menu();
                break;
            //----------------------------------------------------------------
            case 'x': // switch settings
            case 'X':
                switch_settings_menu();
                break;
            //----------------------------------------------------------------
            case 'v': // apply view to frame(s) data
            case 'V':
                apply_view_menu();
                break;
            //----------------------------------------------------------------
            case 'p': // Palette transformations
            case 'P':
                palette_transforms_menu();
                break;
            //----------------------------------------------------------------
            case 'h': // frame transforms
            case 'H':
                frame_transforms_menu();
                break;
            //----------------------------------------------------------------
            case 'j': // frame set transforms
            case 'J':
                frame_set_transforms_menu();
                break;
            //----------------------------------------------------------------
            case 'k': // draw, color, blank
            case 'K':
                draw_color_blank_menu();
                break;
            //----------------------------------------------------------------
            case 'l': // move scale rotate
            case 'L':
                move_scale_rotate_menu();
                break;
            //----------------------------------------------------------------
            case 'b': // wave utilities
            case 'B':
                wave_utilities_menu();
                break;
            //----------------------------------------------------------------
            case 'i': // open data from file
            case 'I':
                file_open_menu();
                break;
            //----------------------------------------------------------------
            case 'o': // save contents of memory to various file types
            case 'O':
                save_as_file_menu();
                break;
        } // end switch(event.key.keysym.unicode)
        //--------------------------------------------------------------------
        if(teminate_requested)
            process_terminate_request();
        //--------------------------------------------------------------------
    } // end while(running)
    //------------------------------------------------------------------------
    Laserboy_terminating();
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
